/*
 * Copyright (C) 2011 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling;

import org.teleal.cling.binding.xml.DeviceDescriptorBinder;
import org.teleal.cling.binding.xml.ServiceDescriptorBinder;
import org.teleal.cling.binding.xml.UDA10DeviceDescriptorBinderImpl;
import org.teleal.cling.binding.xml.UDA10ServiceDescriptorBinderImpl;
import org.teleal.cling.model.ModelUtil;
import org.teleal.cling.model.Namespace;
import org.teleal.cling.model.types.ServiceType;
import org.teleal.cling.transport.impl.DatagramIOConfigurationImpl;
import org.teleal.cling.transport.impl.DatagramIOImpl;
import org.teleal.cling.transport.impl.DatagramProcessorImpl;
import org.teleal.cling.transport.impl.GENAEventProcessorImpl;
import org.teleal.cling.transport.impl.MulticastReceiverConfigurationImpl;
import org.teleal.cling.transport.impl.MulticastReceiverImpl;
import org.teleal.cling.transport.impl.NetworkAddressFactoryImpl;
import org.teleal.cling.transport.impl.SOAPActionProcessorImpl;
import org.teleal.cling.transport.impl.StreamClientConfigurationImpl;
import org.teleal.cling.transport.impl.StreamClientImpl;
import org.teleal.cling.transport.impl.StreamServerConfigurationImpl;
import org.teleal.cling.transport.impl.StreamServerImpl;
import org.teleal.cling.transport.spi.DatagramIO;
import org.teleal.cling.transport.spi.DatagramProcessor;
import org.teleal.cling.transport.spi.GENAEventProcessor;
import org.teleal.cling.transport.spi.MulticastReceiver;
import org.teleal.cling.transport.spi.NetworkAddressFactory;
import org.teleal.cling.transport.spi.SOAPActionProcessor;
import org.teleal.cling.transport.spi.StreamClient;
import org.teleal.cling.transport.spi.StreamServer;
import org.teleal.cling.common.util.Exceptions;
import org.teleal.cling.model.Constants;
import org.teleal.cling.protocol.ProtocolFactory;
import org.teleal.cling.transport.Router;

import java.net.Inet4Address;
import java.net.Inet6Address;
import java.util.concurrent.Executor;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
// import java.util.logging.Logger;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Enumeration;
import java.util.concurrent.LinkedBlockingQueue;
// import java.util.logging.Level;

/**
 * Default configuration data of a typical UPnP stack.
 * <p>
 * This configuration utilizes the default network transport implementation found in
 * {@link org.teleal.cling.transport.impl}.
 * </p>
 * <p>
 * This configuration utilizes the DOM default descriptor binders found in
 * {@link org.teleal.cling.binding.xml}.
 * </p>
 * <p>
 * The thread <code>Executor</code> is an <code>Executors.newCachedThreadPool()</code> with
 * a custom {@link ClingThreadFactory} (it only sets a thread name).
 * </p>
 * <p>
 * Note that this pool is effectively unlimited, so the number of threads will
 * grow (and shrink) as needed - or restricted by your JVM.
 * </p>
 * <p>
 * The default {@link org.teleal.cling.model.Namespace} is configured without any
 * base path or prefix.
 * </p>
 *
 * @author Christian Bauer
 */
public class DefaultUpnpServiceConfiguration implements UpnpServiceConfiguration {

    // private static Logger log = Logger.getLogger(DefaultUpnpServiceConfiguration.class.getName());

    private int streamListenPort=49152;
    private int maxTotalConnections = 1024;    
    private int connectionTimeoutSeconds = 10;
    private int dataReadTimeoutSeconds = 25;
    private int socketbuffersize = 65536;
    private String contentCharset = "UTF-8";    
    public static String productName = Constants.PRODUCT_TOKEN_NAME;
    public static String productVersion = Constants.PRODUCT_TOKEN_VERSION;
    private InetAddress localAddress = null;
    private InetAddress localAddress6 = null;
    private int ttl = 8;
    final private Executor defaultExecutor;
    private StreamClient streamClient = null;
    private StreamServer streamServer = null;
    private static MulticastReceiver multicastReceiver = null;
    private static DatagramIO datagramIO = null;
    private static DatagramProcessor datagramProcessor = null;
    private static SOAPActionProcessor soapActionProcessor = null;
    private static GENAEventProcessor genaEventProcessor = null;
    private int streamBufferSizeKilobytes = 64;
    private Boolean streamTcpNoDelay = true;
    private InetAddress group = null; // = InetAddress.getByName(Constants.IPV4_UPNP_MULTICAST_GROUP);
    private int port = 1900;
    private int maxDatagramBytes = 640;
    final private DeviceDescriptorBinder deviceDescriptorBinderUDA10;
    final private ServiceDescriptorBinder serviceDescriptorBinderUDA10;
    private static Router router = null;
    private Namespace namespace;
    public String baseURL = "";

    /**
     * Defaults to port '0', ephemeral.
     */
    public DefaultUpnpServiceConfiguration() {
        this(NetworkAddressFactoryImpl.DEFAULT_TCP_HTTP_LISTEN_PORT);
    }

    public DefaultUpnpServiceConfiguration(int streamListenPort) {
        this(streamListenPort, true);
    }

    protected DefaultUpnpServiceConfiguration(boolean checkRuntime) {
        this(NetworkAddressFactoryImpl.DEFAULT_TCP_HTTP_LISTEN_PORT, checkRuntime);
    }

    protected DefaultUpnpServiceConfiguration(int streamListenPort, boolean checkRuntime) {
        // if (checkRuntime && ModelUtil.ANDROID_RUNTIME) {
        //     throw new Error("Unsupported runtime environment, use org.teleal.cling.android.AndroidUpnpServiceConfiguration");
        // }

        this.streamListenPort = streamListenPort;

        defaultExecutor = createDefaultExecutor();

        datagramProcessor = createDatagramProcessor();
        soapActionProcessor = createSOAPActionProcessor();
        genaEventProcessor = createGENAEventProcessor();

        deviceDescriptorBinderUDA10 = createDeviceDescriptorBinderUDA10();
        serviceDescriptorBinderUDA10 = createServiceDescriptorBinderUDA10();
        if (localAddress != null) { 
            if (localAddress instanceof Inet6Address) {
              baseURL = "http://[" + localAddress.getHostAddress() + "]:" + this.streamListenPort;
            } else {
              baseURL = "http://" + localAddress.getHostAddress() + ":" + this.streamListenPort;
            } 
        } else {
          if (localAddress6 != null) { 
             baseURL = "http://[" + localAddress6.getHostAddress() + "]:" + this.streamListenPort; 
          } 
        }
        namespace = createNamespace(baseURL);
    }

    @Override
    public DatagramProcessor getDatagramProcessor() {
        return datagramProcessor;
    }

    @Override
    public SOAPActionProcessor getSoapActionProcessor() {
        return soapActionProcessor;
    }

    @Override
    public GENAEventProcessor getGenaEventProcessor() {
        return genaEventProcessor;
    }

    @Override
    public StreamClient createStreamClient() {
        StreamClientConfigurationImpl clientConfig = new StreamClientConfigurationImpl();
        clientConfig.setConnectionTimeoutSeconds(connectionTimeoutSeconds);
        clientConfig.setMaxTotalConnections(maxTotalConnections);
        clientConfig.setDataReadTimeoutSeconds(dataReadTimeoutSeconds);
        clientConfig.setContentCharset(contentCharset);
        clientConfig.setSocketBufferSize(socketbuffersize);
        clientConfig.setProductName(productName);
        clientConfig.setProductVersion(productVersion);
        clientConfig.setRouter(router);
        streamClient = new StreamClientImpl(clientConfig);
        return streamClient;       
     }

    @Override
    public MulticastReceiver createMulticastReceiver(NetworkAddressFactory networkAddressFactory, UpnpServiceConfiguration upnpConfiguration, ProtocolFactory protocolFactory) {
        multicastReceiver = new MulticastReceiverImpl(
                new MulticastReceiverConfigurationImpl(
                        networkAddressFactory.getMulticastGroup(),
                        networkAddressFactory.getMulticastPort()),
                upnpConfiguration,
                protocolFactory
        );
        return multicastReceiver;
    }

    @Override
    public DatagramIO createDatagramIO(NetworkAddressFactory networkAddressFactory, UpnpServiceConfiguration upnpConfiguration, ProtocolFactory protocolFactory) {
        datagramIO = new DatagramIOImpl(new DatagramIOConfigurationImpl(), upnpConfiguration, protocolFactory);
        return datagramIO;
    }

    @Override
    public StreamServer createStreamServer(NetworkAddressFactory networkAddressFactory) {
           streamServer = null;
           StreamServerConfigurationImpl config = new StreamServerConfigurationImpl(streamListenPort);
           if (config != null) {
             config.setListenPort(49152);
             config.setBufferSizeKilobytes(120);
             config.setTcpNoDelay(true);
             config.setRouter(router);
             streamServer = new StreamServerImpl(config);
           }
           return streamServer;
       
    }

    @Override
    public Executor getMulticastReceiverExecutor() {
        return getDefaultExecutor();
    }

    @Override
    public Executor getDatagramIOExecutor() {
        return getDefaultExecutor();
    }

    @Override
    public Executor getStreamServerExecutor() {
        return getDefaultExecutor();
    }

    @Override
    public DeviceDescriptorBinder getDeviceDescriptorBinderUDA10() {
        return deviceDescriptorBinderUDA10;
    }

    @Override
    public ServiceDescriptorBinder getServiceDescriptorBinderUDA10() {
        return serviceDescriptorBinderUDA10;
    }

    @Override
    public ServiceType[] getExclusiveServiceTypes() {
        return new ServiceType[0];
    }

    @Override
    public int getRegistryMaintenanceIntervalMillis() {
        return 1000;
    }

    @Override
    public Executor getAsyncProtocolExecutor() {
        return getDefaultExecutor();
    }

    @Override
    public Executor getSyncProtocolExecutor() {
        return getDefaultExecutor();
    }

    @Override
    public Namespace getNamespace() {
        return namespace;
    }

    @Override
    public Executor getRegistryMaintainerExecutor() {
        return getDefaultExecutor();
    }

    @Override
    public Executor getRegistryListenerExecutor() {
        return getDefaultExecutor();
    }

    @Override
    public NetworkAddressFactory createNetworkAddressFactory() {
        return createNetworkAddressFactory(streamListenPort);
    }

    @Override
    public void shutdown() {
        if (getDefaultExecutor() instanceof ThreadPoolExecutor) {
            // log.fine("Shutting down thread pool");
            ((ThreadPoolExecutor) getDefaultExecutor()).shutdown();
        }
    }

    protected NetworkAddressFactory createNetworkAddressFactory(int streamListenPort) {
        return new NetworkAddressFactoryImpl(streamListenPort);
    }

    protected DatagramProcessor createDatagramProcessor() {
        return new DatagramProcessorImpl();
    }

    protected SOAPActionProcessor createSOAPActionProcessor() {
        return new SOAPActionProcessorImpl();
    }

    protected GENAEventProcessor createGENAEventProcessor() {
        return new GENAEventProcessorImpl();
    }

    protected DeviceDescriptorBinder createDeviceDescriptorBinderUDA10() {
        return new UDA10DeviceDescriptorBinderImpl();
    }

    protected ServiceDescriptorBinder createServiceDescriptorBinderUDA10() {
        return new UDA10ServiceDescriptorBinderImpl();
    }

    protected Namespace createNamespace(String url) {
        return new Namespace(url);
    }

    protected Executor getDefaultExecutor() {
        return defaultExecutor;
    }

    protected Executor createDefaultExecutor() {
        return new ClingExecutor();
    }

    public static class ClingExecutor extends ThreadPoolExecutor {

        public ClingExecutor() {
            this(new ClingThreadFactory(),
                 new ThreadPoolExecutor.DiscardPolicy() {
                     // The pool is unbounded but rejections will happen during shutdown
                     @Override
                     public void rejectedExecution(Runnable runnable, ThreadPoolExecutor threadPoolExecutor) {
                         // Log and discard
                         // log.info("Thread pool rejected execution of " + runnable.getClass());
                         super.rejectedExecution(runnable, threadPoolExecutor);
                     }
                 }
            );
        }

        public ClingExecutor(ThreadFactory threadFactory, RejectedExecutionHandler rejectedHandler) {
            // This is the same as Executors.newCachedThreadPool
            // ThreadPoolExecutor defaultExecutor = new ThreadPoolExecutor(8, 16, 3, TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>()) {  // new ArrayBlockingQueue(8192)
            super(4,
                  64,
                  32L,
                  TimeUnit.SECONDS,
                  new LinkedBlockingQueue<Runnable>(),
                  threadFactory,
                  rejectedHandler
            ); 
            /* super(4,
                  Integer.MAX_VALUE,
                  120L,
                  TimeUnit.SECONDS,
                  new SynchronousQueue<Runnable>(),
                  threadFactory,
                  rejectedHandler
            ); */
        }

        @Override
        protected void afterExecute(Runnable runnable, Throwable throwable) {
            super.afterExecute(runnable, throwable);
            if (throwable != null) {
                // Log only
                // log.warning("Thread terminated " + runnable + " abruptly with exception: " + throwable);
                // log.warning("Root cause: " + Exceptions.unwrap(throwable));
            }
        }
    }

    // Executors.DefaultThreadFactory is package visibility (...no touching, you unworthy JDK user!)
    public static class ClingThreadFactory implements ThreadFactory {

        protected final ThreadGroup group;
        protected final AtomicInteger threadNumber = new AtomicInteger(1);
        protected final String namePrefix = "cling-";

        public ClingThreadFactory() {
            SecurityManager s = System.getSecurityManager();
            group = (s != null) ? s.getThreadGroup() : Thread.currentThread().getThreadGroup();
        }

        public Thread newThread(Runnable r) {
            Thread t = new Thread(
                    group, r,
                    namePrefix + threadNumber.getAndIncrement(),
                    0
            );
            if (t.isDaemon())
                t.setDaemon(false);
            if (t.getPriority() != Thread.NORM_PRIORITY)
                t.setPriority(Thread.NORM_PRIORITY);

            return t;
        }
    }

    @Override
    public void setTcpNoDelay(boolean tcpNoDelay) {
        this.streamTcpNoDelay = tcpNoDelay;
    }    
 
    @Override
    public void setBufferSizeKilobytes(int bufferSizeKilobytes) {
        this.streamBufferSizeKilobytes = bufferSizeKilobytes;
    }
 
    @Override
    public void setListenPort(int listenPort) {
        this.streamListenPort = listenPort;
    }
    
    @Override
    public void setMaxTotalConnections(int maxTotalConnections) {
        this.maxTotalConnections = maxTotalConnections;
    }    

    @Override
    public void setConnectionTimeoutSeconds(int connectionTimeoutSeconds) {
        this.connectionTimeoutSeconds = connectionTimeoutSeconds;
    }    

    @Override
    public void setDataReadTimeoutSeconds(int dataReadTimeoutSeconds) {
        this.dataReadTimeoutSeconds = dataReadTimeoutSeconds;
    }    

    @Override
    public void setContentCharset(String contentCharset) {
        this.contentCharset = contentCharset;
    } 
    
    @Override
    public void setSocketBufferSize(int socketbuffersize) {
        this.socketbuffersize = socketbuffersize;
    }

    @Override
    public void setGroup(InetAddress group) {
        this.group = group;
    }

    @Override
    public void setPort(int port) {
        this.port = port;
    }
    @Override
    public void setTTL(int t) {
        this.ttl = t;
    }
    
    @Override
    public void setMaxDatagramBytes(int maxDatagramBytes) {
        this.maxDatagramBytes = maxDatagramBytes;
    }    
    
    @Override
    public void setProductName(String productName) {
        this.productName = productName;
        // Log.v("AndroidUpnpServiceConfiguration", "setting ProductName:"+productName); 
    }

    @Override
    public void setProductVersion(String productVersion) {
        this.productVersion = productVersion;
        // Log.v("AndroidUpnpServiceConfiguration", "setting ProductVersion:"+productVersion);
    }         

    public static String getProductName() {
        // Log.v("DefaultUpnpServiceConfiguration", "getting ProductName:"+productName);
        return productName;
    }

    public static String getProductVersion() {
        // Log.v("DefaultUpnpServiceConfiguration", "getting ProductVersion:"+productVersion);
        return productVersion;
    }    

    @Override
    public void setLocalInetAddress(InetAddress addr) {
        localAddress = addr;
        if (localAddress != null) { 
            if (localAddress instanceof Inet6Address) {
              baseURL = "http://[" + localAddress.getHostAddress() + "]:" + this.streamListenPort;
            } else {
              baseURL = "http://" + localAddress.getHostAddress() + ":" + this.streamListenPort;
            } 
        } else {
          if (localAddress6 != null) { 
             baseURL = "http://[" + localAddress6.getHostAddress() + "]:" + this.streamListenPort; 
          } 
        }
        namespace = createNamespace(baseURL);        
    }     

    @Override
    public InetAddress getLocalInetAddress() {
        return localAddress;
    } 

    @Override
    public void setLocalInetAddress6(InetAddress addr) {
        this.localAddress6 = addr;
    }     

    @Override
    public InetAddress getLocalInetAddress6() {
        return localAddress6;
    } 
    
    @Override
    public void setRouter(Router routr) {
        this.router = routr;
    }     

    @Override
    public Router getRouter() {
        return router;
    }     
 
    static int byteArrayToInt(byte[] arr, int offset) {
        if (arr == null || arr.length - offset < 4)
            return -1;

        int r0 = (arr[offset] & 0xFF) << 24;
        int r1 = (arr[offset + 1] & 0xFF) << 16;
        int r2 = (arr[offset + 2] & 0xFF) << 8;
        int r3 = arr[offset + 3] & 0xFF;
        return r0 + r1 + r2 + r3;
    }
    
    @Override
    public boolean getTcpNoDelay() { return streamTcpNoDelay; };
    @Override
    public int getBufferSizeKilobytes() { return streamBufferSizeKilobytes; };
    @Override
    public int getListenPort() { return streamListenPort; };
    @Override
    public int getMaxTotalConnections() { return maxTotalConnections; };
    @Override
    public int getConnectionTimeoutSeconds() { return connectionTimeoutSeconds; };
    @Override
    public int getDataReadTimeoutSeconds() { return dataReadTimeoutSeconds; };
    @Override
    public String getContentCharset() { return contentCharset; };
    @Override
    public int getSocketBufferSize() { return socketbuffersize; };
    @Override
    public String getGroup() { return  "239.255.255.250"; };
    @Override
    public int getPort() { return port; };
    @Override
    public int getTTL() { return ttl; };    
    @Override
    public int getMaxDatagramBytes() { return streamListenPort; };
       
}
