/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.logging.Level;
import org.teleal.cling.controlpoint.ControlPoint;
import org.teleal.cling.controlpoint.ControlPointImpl;
import org.teleal.cling.protocol.ProtocolFactory;
import org.teleal.cling.protocol.ProtocolFactoryImpl;
import org.teleal.cling.registry.Registry;
import org.teleal.cling.registry.RegistryImpl;
import org.teleal.cling.registry.RegistryListener;
import org.teleal.cling.transport.Router;
import org.teleal.cling.transport.RouterImpl;

// import java.util.logging.Logger;
import org.teleal.cling.model.Constants;
import org.teleal.cling.transport.impl.StreamClientImpl;
import org.teleal.cling.transport.impl.StreamClientConfigurationImpl;
import org.teleal.cling.transport.impl.StreamServerImpl;
import org.teleal.cling.transport.impl.StreamServerConfigurationImpl;
import org.teleal.cling.transport.impl.MulticastReceiverImpl;
import org.teleal.cling.transport.impl.MulticastReceiverConfigurationImpl;
import org.teleal.cling.transport.spi.DatagramIO;
import org.teleal.cling.transport.spi.DatagramProcessor;
import org.teleal.cling.transport.spi.GENAEventProcessor;
import org.teleal.cling.transport.spi.MulticastReceiver;
import org.teleal.cling.transport.spi.SOAPActionProcessor;
import org.teleal.cling.transport.spi.StreamClient;
import org.teleal.cling.transport.spi.StreamServer;

/**
 * Default implementation of {@link UpnpService}, starts immediately on construction.
 * <p>
 * If no {@link UpnpServiceConfiguration} is provided it will automatically
 * instantiate {@link DefaultUpnpServiceConfiguration}. This configuration <strong>does not
 * work</strong> on Android! Use the {@link org.teleal.cling.android.AndroidUpnpService}
 * application component instead
 * </p>
 * <p>
 * Override the various <tt>create...()</tt> methods to customize instantiation of protocol factory,
 * router, etc.
 * </p>
 *
 * @author Christian Bauer
 */
public class UpnpServiceImpl implements UpnpService {

    // private static Logger log = Logger.getLogger(UpnpServiceImpl.class.getName());

    private UpnpServiceConfiguration configuration;
    private ControlPoint controlPoint;
    private ProtocolFactory protocolFactory = null;
    private Registry registry;
    private static Router router;
    private StreamClient streamClient = null;
    private StreamServer streamServer = null;
    private MulticastReceiver multicastReceiver = null;
    private  StreamClientConfigurationImpl clientconfig = new StreamClientConfigurationImpl();
    private  StreamServerConfigurationImpl serverconfig = new StreamServerConfigurationImpl();
    private InetAddress group = null; 
    private MulticastReceiverConfigurationImpl multicastconfig ;
    private InetAddress addr = null; 
    private DatagramIO datagramIO = null;
    private DatagramProcessor datagramProcessor = null;
    private SOAPActionProcessor soapActionProcessor = null;
    private GENAEventProcessor genaEventProcessor = null;
    
    public UpnpServiceImpl() {
        this(new DefaultUpnpServiceConfiguration());
    }

    public UpnpServiceImpl(RegistryListener... registryListeners) {
        this(new DefaultUpnpServiceConfiguration(), registryListeners);
    }

    public UpnpServiceImpl(UpnpServiceConfiguration configuration, RegistryListener... registryListeners) {
        this.configuration = configuration;

        // Log.v("eXport-it UpnpServiceImpl",">>> Starting UPnP service...");

        // Log.v("eXport-it UpnpServiceImpl","Using configuration: " + getConfiguration().getClass().getName());

        // Instantiation order is important: Router needs to start its network services after registry is ready

        this.protocolFactory = createProtocolFactory();

        this.registry = createRegistry(protocolFactory);
        for (RegistryListener registryListener : registryListeners) {
            this.registry.addListener(registryListener);
        }
        addr = this.configuration.getLocalInetAddress();
        // if (addr == null) { Log.v("eXport-it UpnpServiceImpl","inet address is null"); }
        // this.router = createRouter(protocolFactory, registry, addr);
        
        this.controlPoint = createControlPoint(protocolFactory, registry);
        
        this.datagramProcessor = this.configuration.getDatagramProcessor();
        this.soapActionProcessor = this.configuration.getSoapActionProcessor();
        this.genaEventProcessor = this.configuration.getGenaEventProcessor(); 
        
        // Log.v("eXport-it UpnpServiceImpl","<<< UPnP service started successfully");
        
    }

    protected ProtocolFactory createProtocolFactory() {
        return new ProtocolFactoryImpl(this);
    }

    protected Registry createRegistry(ProtocolFactory protocolFactory) {
        return new RegistryImpl(this);
    }

    protected Router createRouter(ProtocolFactory protocolFactory, Registry registry, InetAddress addr) {
        return new RouterImpl(getConfiguration(), protocolFactory, addr);
    }
    
    protected ControlPoint createControlPoint(ProtocolFactory protocolFactory, Registry registry) {
        return new ControlPointImpl(getConfiguration(), protocolFactory, registry);
    }

    public UpnpServiceConfiguration getConfiguration() {
        return configuration;
    }

    public ControlPoint getControlPoint() {
        return controlPoint;
    }

    public ProtocolFactory getProtocolFactory() {
        return protocolFactory;
    }

    public Registry getRegistry() {
        return registry;
    }

    public Router getRouter() {
        return router;
    }
    
    public void setRouter(Router r) {
        this.router = r;
    }
    
    synchronized public void shutdown() {
        // Well, since java.util.logging has its own shutdown hook, this
        // might actually make it into the log or not...
        // Log.v("eXport-it UpnpServiceImpl",">>> Shutting down UPnP service...");

        getRegistry().shutdown();
        getRouter().shutdown();
        getConfiguration().shutdown();

        // Log.v("eXport-it UpnpServiceImpl","<<< UPnP service shutdown completed");
    }

    @Override
    public StreamClient getStreamClient() {
        return streamClient;
    }

    @Override
    public StreamServer getStreamServer() {
        return streamServer;
    }

    @Override
    public MulticastReceiver getMulticastReceiver() {
        return multicastReceiver;
    }
    
    @Override
    public DatagramIO getDatagramIO() {
        return datagramIO;
    }    
    
    @Override
    public GENAEventProcessor getGENAEventProcessor() {
        return genaEventProcessor;    
    }
    
    @Override
    public DatagramProcessor getDatagramProcessor() {
        return datagramProcessor;
    }

    @Override
    public SOAPActionProcessor getSOAPActionProcessor() {
        return soapActionProcessor;
    }

    
}
