/* fully
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.protocol;

// import android.util.Log;

import org.teleal.cling.UpnpService;
import org.teleal.cling.binding.xml.DescriptorBindingException;
import org.teleal.cling.binding.xml.DeviceDescriptorBinder;
import org.teleal.cling.binding.xml.ServiceDescriptorBinder;
import org.teleal.cling.model.ValidationError;
import org.teleal.cling.model.ValidationException;
import org.teleal.cling.model.message.StreamRequestMessage;
import org.teleal.cling.model.message.StreamResponseMessage;
import org.teleal.cling.model.message.UpnpRequest;
import org.teleal.cling.model.meta.Icon;
import org.teleal.cling.model.meta.RemoteDevice;
import org.teleal.cling.model.meta.RemoteService;
import org.teleal.cling.model.types.ServiceType;
import org.teleal.cling.registry.RegistrationException;
import org.teleal.cling.common.util.Exceptions;
import org.teleal.cling.model.message.UpnpHeaders;
import org.teleal.cling.model.message.header.ApplicationURLHeader;
import org.teleal.cling.model.message.header.UpnpHeader;

import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.Date;
import java.text.SimpleDateFormat;

/**
 * Retrieves all remote device XML descriptors, parses them, creates an immutable device and service metadata graph.
 * <p>
 * This implementation encapsulates all steps which are necessary to create a fully usable and populated
 * device metadata graph of a particular UPnP device. It starts with an unhydrated and typically just
 * discovered {@link org.teleal.cling.model.meta.RemoteDevice}, the only property that has to be available is
 * its {@link org.teleal.cling.model.meta.RemoteDeviceIdentity}.
 * </p>
 * <p>
 * This protocol implementation will then retrieve the device's XML descriptor, parse it, and retrieve and
 * parse all service descriptors until all device and service metadata has been retrieved. The fully
 * hydrated device is then added to the {@link org.teleal.cling.registry.Registry}.
 * </p>
 * <p>
 * Any descriptor retrieval, parsing, or validation error of the metadata will abort this protocol
 * with a warning message in the log.
 * </p>
 *
 * @author Christian Bauer
 */
public class RetrieveRemoteDescriptors implements Runnable {

    // final private static Logger log = Logger.getLogger(RetrieveRemoteDescriptors.class.getName());

    private static UpnpService upnpService = null;
    private RemoteDevice rd = null;

    private static final Set<URL> activeRetrievals = new CopyOnWriteArraySet();

    public RetrieveRemoteDescriptors(UpnpService upnpService, RemoteDevice rd) {
        this.upnpService = upnpService;
        this.rd = rd;
        // Log.v("eXport-it RetrieveRemoteDescriptors","Initializing... ");
    }

    public UpnpService getUpnpService() {
        return upnpService;
    }

    @Override
    public void run() {
      if (rd != null) {
        URL deviceURL = rd.getIdentity().getDescriptorURL();
        String UrlString = deviceURL.toString();
        // Performance optimization, try to avoid concurrent GET requests for device descriptor,
        // if we retrieve it once, we have the hydrated device. There is no different outcome
        // processing this several times concurrently.

        if (activeRetrievals.contains(deviceURL)) {
            // Log.v("eXport-it RetrieveRemoteDescriptors","Exiting early, active retrieval for URL already in progress: " + deviceURL);
            return;
        }
        if (getUpnpService() != null) {
          if (getUpnpService().getRegistry() != null) {  
             // Exit if it has been discovered already, could be have been waiting in the executor queue too long
             if (getUpnpService().getRegistry().getRemoteDevice(rd.getIdentity().getUdn(), true) != null) {
                // Log.v("eXport-it RetrieveRemoteDescriptors","Exiting early, already discovered: " + deviceURL);
                return;
             }
          }
          if (getUpnpService().getConfiguration() != null) {
            if (getUpnpService().getConfiguration().getRouter() != null) {  
              try {
                 activeRetrievals.add(deviceURL);
                 // getUpnpService().getConfiguration().getRouter().getStreamClient()  Null pointer !!!!
                 if (getUpnpService().getConfiguration().getRouter().getStreamClient() != null) { 
                     describe();
                 }
              } finally {
                 activeRetrievals.remove(deviceURL);
              }
            }
          }  
        }
      }  
    }

    protected void describe() {

        // All of the following is a very expensive and time consuming procedure, thanks to the
        // braindead design of UPnP. Several GET requests, several descriptors, several XML parsing
        // steps - all of this could be done with one and it wouldn't make a difference. So every
        // call of this method has to be really necessary and rare.
      if (rd != null) {
        try {   
            StreamRequestMessage deviceDescRetrievalMsg = new StreamRequestMessage(UpnpRequest.Method.GET, rd.getIdentity().getDescriptorURL());
            // Log.v("eXport-it RetrieveRemoteDescriptors","Sending device descriptor retrieval message: " + deviceDescRetrievalMsg);
            URI descriptorURI = deviceDescRetrievalMsg.getUri();
            StreamResponseMessage deviceDescMsg = null;
            if (descriptorURI.getHost() != null) {
                deviceDescMsg = getUpnpService().getConfiguration().getRouter().getStreamClient().sendRequest(deviceDescRetrievalMsg);
            }
            if (deviceDescMsg == null) {
                // Log.v("eXport-it RetrieveRemoteDescriptors","Device descriptor retrieval failed, no response: " + rd.getIdentity().getDescriptorURL());
                return;
            }

            // JCDufourd: added to deal with DIAL device discovery
            UpnpHeaders deviceDescMsgHeaders = deviceDescMsg.getHeaders();
            // Log.v("eXport-it RetrieveRemoteDescriptors","headers:"+deviceDescMsgHeaders.toString());
            if (deviceDescMsgHeaders.containsKey(UpnpHeader.Type.APPLICATIONURL)) {
               // this remote device is a DIAL device, not really UPnP
               // Log.v("eXport-it RetrieveRemoteDescriptors","headers contains Key ApplicationUrl");
               ApplicationURLHeader h = (ApplicationURLHeader)deviceDescMsgHeaders.getFirstHeader(UpnpHeader.Type.APPLICATIONURL);
               rd.setDIALApplicationURL(h.getValue());
            }
            //JCDufourd: end of addition        
        
            if (deviceDescMsg.getOperation().isFailed()) {
                /* SimpleDateFormat sdfTime = new SimpleDateFormat("HH:mm:ss");
                Date now = new Date();
                String strTime = sdfTime.format(now);
                Log.v("eXport-it RetrieveRemoteDescriptors",
                    strTime+" Device descriptor retrieval failed: "
                            + rd.getIdentity().getDescriptorURL() +
                            ", "
                            + deviceDescMsg.getOperation().getResponseDetails()
                    ); */
               return;
            }

            if (!deviceDescMsg.isContentTypeTextUDA()) {
                // Log.v("eXport-it RetrieveRemoteDescriptors","Received device descriptor without or with invalid Content-Type: " + rd.getIdentity().getDescriptorURL());
                // We continue despite the invalid UPnP message because we can still hope to convert the content
            }
            /* SimpleDateFormat sdfTime = new SimpleDateFormat("HH:mm:ss");
            Date now = new Date();
            String strTime = sdfTime.format(now); */
            
            String descriptorXML = deviceDescMsg.getBodyString();
            // descriptorXML = descriptorXML.replace("<UPC/>","<UPC></UPC>");
            describe(descriptorXML);
            
        } catch (Exception ex) {   } 
      }  
    }

    protected void describe(String descriptorXML) {
      if (getUpnpService() != null) {
        boolean notifiedStart = false;
        RemoteDevice describedDevice = null;
        
        // Log.v("eXport-it RetrieveRemoteDescriptors","describe: " + descriptorXML);
        DeviceDescriptorBinder deviceDescriptorBinder = getUpnpService().getConfiguration().getDeviceDescriptorBinderUDA10();
        if (deviceDescriptorBinder != null) {
          try {   
            describedDevice = deviceDescriptorBinder.describe(rd, descriptorXML);

            // Log.v("eXport-it RetrieveRemoteDescriptors","Remote device described (without services) notifying listeners: " + describedDevice);
            notifiedStart = getUpnpService().getRegistry().notifyDiscoveryStart(describedDevice);

            // Log.v("eXport-it RetrieveRemoteDescriptors","Hydrating described device's services: " + describedDevice);
            RemoteDevice hydratedDevice = describeServices(describedDevice);
            if (hydratedDevice == null) {
                // Log.v("eXport-it RetrieveRemoteDescriptors","Device service description failed: " + rd);
                if (notifiedStart)
                    getUpnpService().getRegistry().notifyDiscoveryFailure(
                            describedDevice,
                            new DescriptorBindingException("Device service description failed: " + rd)
                    );
                return;
            } else {
              if (hydratedDevice.getDetails() != null)  {
                 SimpleDateFormat sdfTime = new SimpleDateFormat("HH:mm:ss"); 
                 Date now = new Date();
                 String strTime = sdfTime.format(now);
                 // Log.v("eXport-it RetrieveRemoteDescriptors",strTime+"Device: "+hydratedDevice.getDetails().getFriendlyName());
                 // Log.v("eXport-it RetrieveRemoteDescriptors","Type: "+hydratedDevice.getType());
              } else {
                 // Log.v("eXport-it RetrieveRemoteDescriptors","fully hydrated device has NO Detail !!! "+descriptorXML.length());  
              }
            }

            // Log.v("eXport-it RetrieveRemoteDescriptors","Adding fully hydrated remote device to registry: " + hydratedDevice);
            // The registry will do the right thing: A new root device is going to be added, if it's
            // already present or we just received the descriptor again (because we got an embedded
            // devices' notification), it will simply update the expiration timestamp of the root
            // device.
            getUpnpService().getRegistry().addDevice(hydratedDevice);

          } catch (ValidationException ex) {
            // Log.v("eXport-it RetrieveRemoteDescriptors","Could not validate device model: " + rd);
            for (ValidationError validationError : ex.getErrors()) {
                 // Log.v("eXport-it RetrieveRemoteDescriptors",validationError.toString());
            }
            if (describedDevice != null && notifiedStart)
                getUpnpService().getRegistry().notifyDiscoveryFailure(describedDevice, ex);

          } catch (DescriptorBindingException ex) {
             // Log.v("eXport-it RetrieveRemoteDescriptors","Could not hydrate device or its services from descriptor: " + rd);
             // Log.v("eXport-it RetrieveRemoteDescriptors","Cause was: " + Exceptions.unwrap(ex));
            if (describedDevice != null && notifiedStart)
                getUpnpService().getRegistry().notifyDiscoveryFailure(describedDevice, ex);

          } catch (RegistrationException ex) {
             // Log.v("eXport-it RetrieveRemoteDescriptors","Adding hydrated device to registry failed: " + rd);
             // Log.v("eXport-it RetrieveRemoteDescriptors","Cause was: " + ex.toString());
            if (describedDevice != null && notifiedStart)
                getUpnpService().getRegistry().notifyDiscoveryFailure(describedDevice, ex);
          } catch (Exception exx) {   }
        } else {
           // Log.v("eXport-it RetrieveRemoteDescriptors","deviceDescriptorBinder is NULL ");  
        } 
      }  
    }

    protected RemoteDevice describeServices(RemoteDevice currentDevice)
            throws DescriptorBindingException, ValidationException {
        if (currentDevice == null) {
            return null;
        }
        List<RemoteService> describedServices = new ArrayList();
        if (currentDevice.hasServices()) {
            List<RemoteService> filteredServices = filterExclusiveServices(currentDevice.getServices());
            for (RemoteService service : filteredServices) {
                RemoteService svc = describeService(service);
                if (svc == null) { // Something went wrong, bail out
                    return null;
                }
                // Log.v("eXport-it RetrieveRemoteDescriptors","RemoteService: "+svc.getServiceType()); 
                describedServices.add(svc);
            }
        }

        List<RemoteDevice> describedEmbeddedDevices = new ArrayList();
        if (currentDevice.hasEmbeddedDevices()) {
            for (RemoteDevice embeddedDevice : currentDevice.getEmbeddedDevices()) {
                if (embeddedDevice == null) continue;
                // Log.v("eXport-it RetrieveRemoteDescriptors","embeddedDevice: "+embeddedDevice.getType().getType());               
                RemoteDevice describedEmbeddedDevice = describeServices(embeddedDevice);
                if (describedEmbeddedDevice == null) { // Something was wrong, recursively
                    // Log.v("eXport-it RetrieveRemoteDescriptors","describeEmbeddedDevice is null "); 
                    return null;
                }
                describedEmbeddedDevices.add(describedEmbeddedDevice);
            }
        }

        Icon[] iconDupes = new Icon[currentDevice.getIcons().length];
        for (int i = 0; i < currentDevice.getIcons().length; i++) {
            Icon icon = currentDevice.getIcons()[i];
            iconDupes[i] = icon.deepCopy();
        }

        // Yes, we create a completely new immutable graph here
        return currentDevice.newInstance(
                currentDevice.getIdentity().getUdn(),
                currentDevice.getVersion(),
                currentDevice.getType(),
                currentDevice.getDetails(),
                iconDupes,
                currentDevice.toServiceArray(describedServices),
                describedEmbeddedDevices
        );
    }

    protected RemoteService describeService(RemoteService service)
            throws DescriptorBindingException, ValidationException {
        if (service == null) {
            return null;
        } else {
            if (service.getDevice() == null) {
                return null;
            }
        }
        if (getUpnpService().getConfiguration() == null) {
            return null;
        } else {
                if (getUpnpService().getConfiguration().getRouter() == null) {
                   return null;
                } else {
                    if (getUpnpService().getConfiguration().getRouter().getStreamClient() == null) {
                      return null;
                    }
                }
        }      
        
        URL descriptorURL = service.getDevice().normalizeURI(service.getDescriptorURI());
        StreamRequestMessage serviceDescRetrievalMsg = null;
        if (descriptorURL.getHost() != null) {
           serviceDescRetrievalMsg = new StreamRequestMessage(UpnpRequest.Method.GET, descriptorURL);
        }
        
        StreamResponseMessage serviceDescMsg = null;
        if (serviceDescRetrievalMsg != null) {
           try {
               // Log.v("eXport-it RetrieveRemoteDescriptors","Sending service descriptor retrieval message: " + serviceDescRetrievalMsg);
               serviceDescMsg = getUpnpService().getConfiguration().getRouter().getStreamClient().sendRequest(serviceDescRetrievalMsg);
           } catch (Exception ex) {    }
        }
        if (serviceDescMsg == null) {
            // Log.v("eXport-it RetrieveRemoteDescriptors","Could not retrieve service descriptor: " + service);
            return null;
        }

        if (serviceDescMsg.getOperation().isFailed()) {
            /* Log.v("eXport-it RetrieveRemoteDescriptors","Service descriptor retrieval failed: "
                                + descriptorURL
                                + ", "
                                + serviceDescMsg.getOperation().getResponseDetails());  
            return null; */
        }

        if (!serviceDescMsg.isContentTypeTextUDA()) {
            // Log.v("eXport-it RetrieveRemoteDescriptors","Received service descriptor without or with invalid Content-Type: " + descriptorURL);
            // We continue despite the invalid UPnP message because we can still hope to convert the content
        }

        String descriptorContent = serviceDescMsg.getBodyString();
        if (descriptorContent == null || descriptorContent.length() == 0) {
            // Log.v("eXport-it RetrieveRemoteDescriptors","Received empty descriptor:" + descriptorURL);
            return null;
        }

        // Log.v("eXport-it RetrieveRemoteDescriptors","Received service descriptor, hydrating service model: " + serviceDescMsg);
        ServiceDescriptorBinder serviceDescriptorBinder =
                getUpnpService().getConfiguration().getServiceDescriptorBinderUDA10();

        return serviceDescriptorBinder.describe(service, serviceDescMsg.getBodyString());
    }

    protected List<RemoteService> filterExclusiveServices(RemoteService[] services) {
        ServiceType[] exclusiveTypes = null; // getUpnpService().getConfiguration().getExclusiveServiceTypes();

        if (exclusiveTypes == null || exclusiveTypes.length == 0)
            return Arrays.asList(services);

        List<RemoteService> exclusiveServices = new ArrayList();
        for (RemoteService discoveredService : services) {
            for (ServiceType exclusiveType : exclusiveTypes) {
                if (discoveredService.getServiceType().implementsVersion(exclusiveType)) {
                    // Log.v("eXport-it RetrieveRemoteDescriptors","Including exlusive service: " + discoveredService);
                    exclusiveServices.add(discoveredService);
                } else {
                    // Log.v("eXport-it RetrieveRemoteDescriptors","Excluding unwanted service: " + exclusiveType);
                }
            }
        }
        return exclusiveServices;
    }

}
