
/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.protocol.async;

// import android.util.Log;

import org.teleal.cling.UpnpService;
import org.teleal.cling.model.Location;
import org.teleal.cling.model.NetworkAddress;
import org.teleal.cling.model.message.IncomingDatagramMessage;
import org.teleal.cling.model.message.UpnpRequest;
import org.teleal.cling.model.message.discovery.IncomingSearchRequest;
import org.teleal.cling.model.message.discovery.OutgoingSearchResponse;
import org.teleal.cling.model.message.discovery.OutgoingSearchResponseDeviceType;
import org.teleal.cling.model.message.discovery.OutgoingSearchResponseRootDevice;
import org.teleal.cling.model.message.discovery.OutgoingSearchResponseRootDeviceUDN;
import org.teleal.cling.model.message.discovery.OutgoingSearchResponseServiceType;
import org.teleal.cling.model.message.discovery.OutgoingSearchResponseUDN;
import org.teleal.cling.model.message.header.DeviceTypeHeader;
import org.teleal.cling.model.message.header.MXHeader;
import org.teleal.cling.model.message.header.RootDeviceHeader;
import org.teleal.cling.model.message.header.STAllHeader;
import org.teleal.cling.model.message.header.ServiceTypeHeader;
import org.teleal.cling.model.message.header.UDNHeader;
import org.teleal.cling.model.message.header.UpnpHeader;
import org.teleal.cling.model.meta.Device;
import org.teleal.cling.model.meta.LocalDevice;
import org.teleal.cling.model.types.DeviceType;
import org.teleal.cling.model.types.ServiceType;
import org.teleal.cling.model.types.UDN;
import org.teleal.cling.protocol.ReceivingAsync;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Random;
import java.util.logging.Logger;

/**
 * Handles reception of search requests, responds for local registered devices.
 * <p>
 * Waits a random time between 0 and the requested <em>MX</em> (maximum 120 seconds)
 * before executing. Only waits if there are actually any registered local devices.
 * </p>
 * <p>
 * Extracts the <em>search target</em>, builds and sends the dozens of messages
 * required by the UPnP specification, depending on the search target and what
 * local devices and services are found in the {@link org.teleal.cling.registry.Registry}.
 * </p>
 *
 * @author Christian Bauer
 */
public class ReceivingSearch extends ReceivingAsync<IncomingSearchRequest> {

    final protected Random randomGenerator = new Random();

    public ReceivingSearch(UpnpService upnpService, IncomingDatagramMessage<UpnpRequest> inputMessage) {
        super(upnpService, new IncomingSearchRequest(inputMessage));
    }

    protected void execute() {
        // Log.v("eXport-it ReceivingSearch","starting execute");
        if (getUpnpService() == null) {
          // Log.v("eXport-it SendingNotification","UpnpService is null ");
          return;
        } else {
          if (getUpnpService().getConfiguration().getRouter() == null) {
             // Log.v("eXport-it SendingNotification","Router is null ");
            return;  
          } 
        } 

        if (!getInputMessage().isMANSSDPDiscover()) {
            // Log.v("eXport-it ReceivingSearch","Invalid search request, no or invalid MAN ssdp:discover header: " + getInputMessage());
            return;
        }

        UpnpHeader searchTarget = getInputMessage().getSearchTarget();

        if (searchTarget == null) {
            // Log.v("eXport-it ReceivingSearch","Invalid search request, did not contain ST header: " + getInputMessage());
            return;
        }

        List<NetworkAddress> activeStreamServers =
            getUpnpService().getConfiguration().getRouter().getActiveStreamServers(getInputMessage().getLocalAddress());
        if (activeStreamServers.size() == 0) {
            // Log.v("eXport-it ReceivingSearch","Aborting search response, no active stream servers found (network disabled?)");
            return;
        }

        for (NetworkAddress activeStreamServer : activeStreamServers) {
            // Log.v("eXport-it ReceivingSearch","sending search response active stream server "+activeStreamServer.getAddress().toString()+":"+activeStreamServer.getPort());
            sendResponses(searchTarget, activeStreamServer);
        }
    }

    @Override
    protected boolean waitBeforeExecution() throws InterruptedException {

        Integer mx = getInputMessage().getMX();

        if (mx == null) {
            // Log.v("eXport-it ReceivingSearch","Invalid search request, did not contain MX header: " + getInputMessage());
            return false;
        }

        // Spec says we should assume "less" if it's 120 or more
        // From the spec, MX should be "greater than or equal to 1"
        // Prevent negative MX to make nextInt() throw IllegalArgumentException below
        if (mx > 120 || mx <= 0) mx = MXHeader.DEFAULT_VALUE;

        // Only wait if there is something to wait for
        if (getUpnpService().getRegistry().getLocalDevices().size() > 0) {
            int sleepTime = randomGenerator.nextInt(mx * 1000);
            // Log.v("eXport-it ReceivingSearch","Sleeping " + sleepTime + " milliseconds to avoid flooding with search responses");
            Thread.sleep(sleepTime);
        } else {
           // Log.v("eXport-it ReceivingSearch","No Local Device"); 
        }

        return true;
    }

    protected void sendResponses(UpnpHeader searchTarget, NetworkAddress activeStreamServer) {
        if (searchTarget instanceof STAllHeader) {

            sendSearchResponseAll(activeStreamServer);

        } else if (searchTarget instanceof RootDeviceHeader) {

            sendSearchResponseRootDevices(activeStreamServer);

        } else if (searchTarget instanceof UDNHeader) {

            sendSearchResponseUDN((UDN) searchTarget.getValue(), activeStreamServer);

        } else if (searchTarget instanceof DeviceTypeHeader) {

            sendSearchResponseDeviceType((DeviceType) searchTarget.getValue(), activeStreamServer);

        } else if (searchTarget instanceof ServiceTypeHeader) {

            sendSearchResponseServiceType((ServiceType) searchTarget.getValue(), activeStreamServer);

        } else {
             // Log.v("eXport-it ReceivingSearch","Non-implemented search request target: " + searchTarget.getClass());
        }
    }

    protected void sendSearchResponseAll(NetworkAddress activeStreamServer) {
        // Log.v("eXport-it ReceivingSearch","Responding to 'all' search with advertisement messages for all local devices");
        if (getUpnpService() == null) {
          // Log.v("eXport-it SendingNotification","UpnpService is null ");
          return;
        } else {
          if ((getUpnpService().getConfiguration().getRouter() == null) || (getUpnpService().getRegistry() == null)) {
            // Log.v("eXport-it SendingNotification","Router is null ");
            return;  
          } else {
              if (getUpnpService().getRegistry().getLocalDevices() == null) {
                 return;   
              }
          }
        } 
        for (LocalDevice localDevice : getUpnpService().getRegistry().getLocalDevices()) {

            // We are re-using the regular notification messages here but override the NT with the ST header

            // Log.v("eXport-it ReceivingSearch","Sending root device messages: " + localDevice);
            // Log.v("eXport-it ReceivingSearch","localDevice baseURL: " + localDevice.getDetails().getBaseURL().toString());
            // Log.v("eXport-it ReceivingSearch","localDevice presentationURI: " + localDevice.getDetails().getPresentationURI().toString());
            List<OutgoingSearchResponse> rootDeviceMsgs =                       // OutgoingSearchResponse
                    createDeviceMessages(localDevice, activeStreamServer);
            for (OutgoingSearchResponse upnpMessage : rootDeviceMsgs) {
                 // Log.v("eXport-it ReceivingSearch","Sending a root device message");
              if (getUpnpService() != null) {  
                if (getUpnpService().getConfiguration().getRouter() != null) {
                  if (getUpnpService().getConfiguration().getRouter().getDatagramIO() != null) {    
                    getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(upnpMessage);
                  }
                }
              }  
            }

            if (localDevice.hasEmbeddedDevices()) {
                for (LocalDevice embeddedDevice : localDevice.findEmbeddedDevices()) {
                    // Log.v("eXport-it ReceivingSearch","Sending embedded device messages: " + embeddedDevice);
                    List<OutgoingSearchResponse> embeddedDeviceMsgs =
                            createDeviceMessages(embeddedDevice, activeStreamServer);
                    for (OutgoingSearchResponse upnpMessage : embeddedDeviceMsgs) {
                       if (getUpnpService() != null) {  
                         if (getUpnpService().getConfiguration().getRouter() != null) {
                           if (getUpnpService().getConfiguration().getRouter().getDatagramIO() != null) {   
                             getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(upnpMessage);
                           }
                         }
                       }  
                    }
                }
            }

            List<OutgoingSearchResponse> serviceTypeMsgs =
                    createServiceTypeMessages(localDevice, activeStreamServer);
            if (serviceTypeMsgs.size() > 0) {
              //  Log.v("eXport-it ReceivingSearch","Sending service type messages");
              try {  
                for (OutgoingSearchResponse upnpMessage : serviceTypeMsgs) {
                  if (getUpnpService() != null) {  
                    if (getUpnpService().getConfiguration().getRouter() != null) {
                      if (getUpnpService().getConfiguration().getRouter().getDatagramIO() != null) {    
                        getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(upnpMessage);
                      }
                    }
                  }                   
                }
              } catch(NullPointerException e) { }  
            }

        }
    }

    protected List<OutgoingSearchResponse> createDeviceMessages(LocalDevice device,
                                                                NetworkAddress activeStreamServer) {
        List<OutgoingSearchResponse> msgs = new ArrayList();

        // See the tables in UDA 1.0 section 1.1.2

        if (device.isRoot()) {
            msgs.add(
                    new OutgoingSearchResponseRootDevice(
                            getInputMessage(),
                            getDescriptorLocation(activeStreamServer, device),
                            device
                    )
            );
        }

        msgs.add(
                new OutgoingSearchResponseUDN(
                        getInputMessage(),
                        getDescriptorLocation(activeStreamServer, device),
                        device
                )
        );

        msgs.add(
                new OutgoingSearchResponseDeviceType(
                        getInputMessage(),
                        getDescriptorLocation(activeStreamServer, device),
                        device
                )
        );

        return msgs;
    }

    protected List<OutgoingSearchResponse> createServiceTypeMessages(LocalDevice device,
                                                                     NetworkAddress activeStreamServer) {
        List<OutgoingSearchResponse> msgs = new ArrayList();
        for (ServiceType serviceType : device.findServiceTypes()) {
            msgs.add(
                    new OutgoingSearchResponseServiceType(
                            getInputMessage(),
                            getDescriptorLocation(activeStreamServer, device),
                            device,
                            serviceType
                    )
            );
        }
        return msgs;
    }

    protected void sendSearchResponseRootDevices(NetworkAddress activeStreamServer) {
        // Log.v("eXport-it ReceivingSearch","Responding to root device search with advertisement messages for all local root devices");
        if (getUpnpService() == null) {
          // Log.v("eXport-it SendingNotification","UpnpService is null ");
          return;
        } else {
          if (getUpnpService().getConfiguration() == null) {  
             return; 
          } else {
            if (getUpnpService().getConfiguration().getRouter() == null) {
               // Log.v("eXport-it SendingNotification","Router is null ");
               return;  
            } else {
                if (getUpnpService().getConfiguration().getRouter().getDatagramIO() == null) {
                   // Log.v("eXport-it SendingNotification","Router is null ");
                   return;  
                } else {
                   if (getUpnpService().getRegistry() == null) {  
                      return; 
                   } else {
                      if (getUpnpService().getRegistry().getLocalDevices() == null) {  
                         return; 
                      } 
                   } 
                }
            }
          }
        } 
        for (LocalDevice device : getUpnpService().getRegistry().getLocalDevices()) {

            getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(
                    new OutgoingSearchResponseRootDeviceUDN(
                            getInputMessage(),
                            getDescriptorLocation(activeStreamServer, device),
                            device
                    )
            );
        }
    }

    protected void sendSearchResponseUDN(UDN udn, NetworkAddress activeStreamServer) {
        if (getUpnpService().getConfiguration() == null) {
          // Log.v("eXport-it SendingNotification","UpnpService is null ");
          return;
        } else {
          if (getUpnpService().getConfiguration().getRouter() == null) {
            //  Log.v("eXport-it SendingNotification","Router is null ");
            return;  
          } 
        } 
        Device device = getUpnpService().getRegistry().getDevice(udn, false);
        if (device != null && device instanceof LocalDevice) {
            // Log.v("eXport-it ReceivingSearch","Responding to UDN device search: " + udn);
            getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(
                    new OutgoingSearchResponseUDN(
                            getInputMessage(),
                            getDescriptorLocation(activeStreamServer, (LocalDevice) device),
                            (LocalDevice) device
                    )
            );
        }
    }

    protected void sendSearchResponseDeviceType(DeviceType deviceType, NetworkAddress activeStreamServer) {
        if (getUpnpService().getConfiguration() == null) {
          // Log.v("eXport-it SendingNotification","UpnpService is null ");
          return;
        } else {
          if (getUpnpService().getConfiguration().getRouter() == null) {
             // Log.v("eXport-it SendingNotification","Router is null ");
            return;  
          } 
        } 
        // Log.v("eXport-it ReceivingSearch","Responding to device type search: " + deviceType);
        Collection<Device> devices = getUpnpService().getRegistry().getDevices(deviceType);
        for (Device device : devices) {
            if (device instanceof LocalDevice) {
              //  Log.v("eXport-it ReceivingSearch","Sending matching device type search result for: " + device);
              if (getUpnpService() != null) {  
                if (getUpnpService().getConfiguration().getRouter() != null) {
                  if (getUpnpService().getConfiguration().getRouter().getDatagramIO() != null) {    
                    getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(
                        new OutgoingSearchResponseDeviceType(
                                getInputMessage(),
                                getDescriptorLocation(activeStreamServer, (LocalDevice) device),
                                (LocalDevice) device
                        )
                    );
                  }
                }
              }  
            }
        }
    }

    protected void sendSearchResponseServiceType(ServiceType serviceType, NetworkAddress activeStreamServer) {
        if (getUpnpService().getConfiguration() == null) {
          // Log.v("eXport-it SendingNotification","UpnpService is null ");
          return;
        } else {
          if (getUpnpService().getConfiguration().getRouter() == null) {
             // Log.v("eXport-it SendingNotification","Router is null ");
            return;  
          } 
        } 
        // Log.v("eXport-it ReceivingSearch","Responding to service type search: " + serviceType);
        Collection<Device> devices = getUpnpService().getRegistry().getDevices(serviceType);
        for (Device device : devices) {
            if (device instanceof LocalDevice) {
              //  Log.v("eXport-it ReceivingSearch","Sending matching service type search result: " + device);
              if (getUpnpService() != null) {  
                if (getUpnpService().getConfiguration().getRouter() != null) {
                  if (getUpnpService().getConfiguration().getRouter().getDatagramIO() != null) {    
                    getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(
                        new OutgoingSearchResponseServiceType(
                                getInputMessage(),
                                getDescriptorLocation(activeStreamServer, (LocalDevice) device),
                                (LocalDevice) device,
                                serviceType
                        )
                    );
                  }
                }
              }  
            }
        }
    }

    protected Location getDescriptorLocation(NetworkAddress activeStreamServer, LocalDevice device) {
        return new Location(
                activeStreamServer,
                getUpnpService().getConfiguration().getNamespace().getDescriptorPath(device)
        );
    }

}