/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.protocol.async;

import java.net.InetAddress;
import java.net.UnknownHostException;
import org.teleal.cling.UpnpService;
import org.teleal.cling.model.Location;
import org.teleal.cling.model.NetworkAddress;
import org.teleal.cling.model.message.discovery.OutgoingNotificationRequest;
import org.teleal.cling.model.message.discovery.OutgoingNotificationRequestDeviceType;
import org.teleal.cling.model.message.discovery.OutgoingNotificationRequestRootDevice;
import org.teleal.cling.model.message.discovery.OutgoingNotificationRequestServiceType;
import org.teleal.cling.model.message.discovery.OutgoingNotificationRequestUDN;
import org.teleal.cling.model.meta.LocalDevice;
import org.teleal.cling.model.types.NotificationSubtype;
import org.teleal.cling.model.types.ServiceType;
import org.teleal.cling.protocol.SendingAsync;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Sending notification messages for a registered local device.
 * <p>
 * Sends all required (dozens) of messages three times, waits between 0 and 150
 * milliseconds between each bulk sending procedure.
 * </p>
 *
 * @author Christian Bauer
 */
public abstract class SendingNotification extends SendingAsync {
    volatile boolean running = true;
    private LocalDevice device;

    public SendingNotification(UpnpService upnpService, LocalDevice device) {
        super(upnpService);
        // Log.v("eXport-it SendingNotification","upnpService: "+upnpService+"  device: "+device);
        this.device = device;
    }

    public LocalDevice getDevice() {
        return device;
    }

    protected void execute() {
        // Log.v("eXport-it SendingNotification","starting execute ");
        running = true;
        if (getUpnpService() == null) {
          // Log.v("eXport-it SendingNotification","UpnpService is null ");
          return;
        } else {
          if (getUpnpService().getConfiguration().getRouter() == null) {
             // Log.v("eXport-it SendingNotification","Router is null ");
            return;  
          } else {
          } 
        } 
  
        List<NetworkAddress> activeStreamServers =
            getUpnpService().getConfiguration().getRouter().getActiveStreamServers(null);
        if (activeStreamServers.size() == 0) {
            // Log.v("eXport-it SendingNotification","Aborting notifications, no active stream servers found (network disabled?)");
            return;
        } 
        // Log.v("eXport-it SendingNotification","before descriptorLocations");   
        // Prepare it once, it's the same for each repetition
        List<Location> descriptorLocations = new ArrayList();
        for (NetworkAddress activeStreamServer : activeStreamServers) {
            descriptorLocations.add(
                    new Location(
                            activeStreamServer,
                            getUpnpService().getConfiguration().getNamespace().getDescriptorPath(getDevice())
                    )
            );
        }
        // Log.v("eXport-it SendingNotification","after descriptorLocations");
        for (int i = 0; i < getBulkRepeat(); i++) {
            try {

                for (Location descriptorLocation : descriptorLocations) {
                    sendMessages(descriptorLocation);
                }

                // UDA 1.0 is silent about this but UDA 1.1 recomments "a few hundred milliseconds"
                // Log.v("eXport-it SendingNotification","Sleeping " + getBulkIntervalMilliseconds() + " milliseconds");
                Thread.sleep(getBulkIntervalMilliseconds());

            } catch (InterruptedException ex) {
                // Log.v("eXport-it SendingNotification","Advertisement thread was interrupted: " + ex);
                return;
            }
        }
    }

    protected int getBulkRepeat() {
        return 3; // UDA 1.0 says maximum 3 times for alive messages, let's just do it for all
    }

    protected int getBulkIntervalMilliseconds() {
        return 150;
    }

    public void sendMessages(Location descriptorLocation) {
        // Log.v("eXport-it SendingNotification","Sending root device messages: " + getDevice());
      if (getUpnpService().getConfiguration().getRouter() != null) {
       if (getUpnpService().getConfiguration().getRouter().getDatagramIO() != null) {   
        List<OutgoingNotificationRequest> rootDeviceMsgs =
                createDeviceMessages(getDevice(), descriptorLocation);
        for (OutgoingNotificationRequest upnpMessage : rootDeviceMsgs) {
            // Log.v("eXport-it SendingNotification","upnpMessage dest: "+upnpMessage.getDestinationAddress().toString()+"  port: "+upnpMessage.getDestinationPort());
            getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(upnpMessage);
        }
 
        if (getDevice().hasEmbeddedDevices()) {
            for (LocalDevice embeddedDevice : getDevice().findEmbeddedDevices()) {
                // Log.v("eXport-it SendingNotification","Sending embedded device messages: " + embeddedDevice);
                List<OutgoingNotificationRequest> embeddedDeviceMsgs =
                        createDeviceMessages(embeddedDevice, descriptorLocation);
                for (OutgoingNotificationRequest upnpMessage : embeddedDeviceMsgs) {
                  if (getUpnpService() != null) {  
                    if (getUpnpService().getConfiguration().getRouter() != null) {
                      if (getUpnpService().getConfiguration().getRouter().getDatagramIO() != null) {    
                        getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(upnpMessage);
                      }
                    }
                  }  
                }
            }
        }

        List<OutgoingNotificationRequest> serviceTypeMsgs =
                createServiceTypeMessages(getDevice(), descriptorLocation);
        if (serviceTypeMsgs.size() > 0) {
            // Log.v("eXport-it SendingNotification","Sending service type messages");
            for (OutgoingNotificationRequest upnpMessage : serviceTypeMsgs) {
              if (getUpnpService() != null) {  
                if (getUpnpService().getConfiguration().getRouter() != null) {
                  if (getUpnpService().getConfiguration().getRouter().getDatagramIO() != null) {  
                    getUpnpService().getConfiguration().getRouter().getDatagramIO().sendAsync(upnpMessage);
                  }
                }   
              }  
            }
        }
       } 
      }  
    }

    protected List<OutgoingNotificationRequest> createDeviceMessages(LocalDevice device,
                                                                     Location descriptorLocation) {
        List<OutgoingNotificationRequest> msgs = new ArrayList();

        // See the tables in UDA 1.0 section 1.1.2

        if (device.isRoot()) {
            msgs.add(
                    new OutgoingNotificationRequestRootDevice(
                            descriptorLocation,
                            device,
                            getNotificationSubtype()
                    )
            );
        }

        msgs.add(
                new OutgoingNotificationRequestUDN(
                        descriptorLocation, device, getNotificationSubtype()
                )
        );
        msgs.add(
                new OutgoingNotificationRequestDeviceType(
                        descriptorLocation, device, getNotificationSubtype()
                )
        );

        return msgs;
    }

    protected List<OutgoingNotificationRequest> createServiceTypeMessages(LocalDevice device,
                                                                          Location descriptorLocation) {
        List<OutgoingNotificationRequest> msgs = new ArrayList();

        for (ServiceType serviceType : device.findServiceTypes()) {
            msgs.add(
                    new OutgoingNotificationRequestServiceType(
                            descriptorLocation, device,
                            getNotificationSubtype(), serviceType
                    )
            );
        }

        return msgs;
    }

    protected abstract NotificationSubtype getNotificationSubtype();

}
