/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.protocol.sync;

import org.teleal.cling.UpnpService;
import org.teleal.cling.model.action.ActionException;
import org.teleal.cling.model.action.ActionInvocation;
import org.teleal.cling.model.message.StreamRequestMessage;
import org.teleal.cling.model.message.StreamResponseMessage;
import org.teleal.cling.model.message.UpnpHeaders;
import org.teleal.cling.model.message.UpnpResponse;
import org.teleal.cling.model.message.control.IncomingActionRequestMessage;
import org.teleal.cling.model.message.control.OutgoingActionResponseMessage;
import org.teleal.cling.model.message.header.ContentTypeHeader;
import org.teleal.cling.model.message.header.UpnpHeader;
import org.teleal.cling.model.resource.ServiceControlResource;
import org.teleal.cling.model.types.ErrorCode;
import org.teleal.cling.protocol.ReceivingSync;
import org.teleal.cling.transport.spi.UnsupportedDataException;
import org.teleal.cling.common.util.Exceptions;

// import android.util.Log;

/**
 * Handles reception of control messages, invoking actions on local services.
 * <p>
 * Actions are invoked through the {@link org.teleal.cling.model.action.ActionExecutor} returned
 * by the registered {@link org.teleal.cling.model.meta.LocalService#getExecutor(org.teleal.cling.model.meta.Action)}
 * method.
 * </p>
 * <p>
 * This class offers two shortcut thread-local variables, which providers of UPnP services might
 * find useful in some situations. You can access these methods statically from within your
 * service implementation:
 * </p>
 * <ul>
 * <li>The {@link #getRequestMessage()} static method offers access to the original action request
 * message, including all received HTTP headers, etc.
 * </li>
 * <li>
 * The {@link #getExtraResponseHeaders()} static method offers modifiable HTTP headers which will
 * be added to the action response after the invocation, and returned to the client.
 * </li>
 * </ul>
 *
 * @author Christian Bauer
 */
public class ReceivingAction extends ReceivingSync<StreamRequestMessage, StreamResponseMessage> {

    final protected static ThreadLocal<IncomingActionRequestMessage> requestThreadLocal = new ThreadLocal();
    final protected static ThreadLocal<UpnpHeaders> extraResponseHeadersThreadLocal = new ThreadLocal();

    public ReceivingAction(UpnpService upnpService, StreamRequestMessage inputMessage) {
        super(upnpService, inputMessage);
    }

    protected StreamResponseMessage executeSync() {

        ContentTypeHeader contentTypeHeader =
                getInputMessage().getHeaders().getFirstHeader(UpnpHeader.Type.CONTENT_TYPE, ContentTypeHeader.class);

        // Special rules for action messages! UDA 1.0 says:
        // 'If the CONTENT-TYPE header specifies an unsupported value (other then "text/xml") the
        // device must return an HTTP status code "415 Unsupported Media Type".'
        // Log.v("eXport-it SyncReceivingAction","Received  Content-Type: '" + contentTypeHeader);
        if (contentTypeHeader != null && !contentTypeHeader.isUDACompliantXML()) {
            // Log.v("eXport-it SyncReceivingAction","Received invalid Content-Type '" + contentTypeHeader + "': " + getInputMessage());
            return new StreamResponseMessage(new UpnpResponse(UpnpResponse.Status.UNSUPPORTED_MEDIA_TYPE));
        }

        /* if (contentTypeHeader == null) {
            Log.v("eXport-it SyncReceivingAction","Received without Content-Type: " + getInputMessage());
        } */

        ServiceControlResource resource =
                getUpnpService().getRegistry().getResource(
                        ServiceControlResource.class,
                        getInputMessage().getUri()
                );

        if (resource == null) {
            // Log.v("eXport-it SyncReceivingAction","No local resource found: " + getInputMessage());
            return null;
        }

        // Log.v("eXport-it SyncReceivingAction","Found local action resource matching relative request URI: " + getInputMessage().getUri());

        ActionInvocation invocation;
        OutgoingActionResponseMessage responseMessage = null;

        try {

            // Throws ActionException if the action can't be found
            IncomingActionRequestMessage requestMessage =
                    new IncomingActionRequestMessage(getInputMessage(), resource.getModel());

            // Preserve message in a TL
            requestThreadLocal.set(requestMessage);
            extraResponseHeadersThreadLocal.set(new UpnpHeaders());

            // Log.v("eXport-it SyncReceivingAction","Created incoming action request message: " + requestMessage);
            invocation = new ActionInvocation(requestMessage.getAction());

            // Throws UnsupportedDataException if the body can't be read
            // Log.v("eXport-it SyncReceivingAction","Reading body of request message");
            getUpnpService().getConfiguration().getSoapActionProcessor().readBody(requestMessage, invocation);

            // Log.v("eXport-it SyncReceivingAction","Executing on local service: " + invocation);
            resource.getModel().getExecutor(invocation.getAction()).execute(invocation);

            if (invocation.getFailure() == null) {
                responseMessage =
                        new OutgoingActionResponseMessage(invocation.getAction());
            } else {
                responseMessage =
                        new OutgoingActionResponseMessage(UpnpResponse.Status.INTERNAL_SERVER_ERROR, invocation.getAction());

            }

        } catch (ActionException ex) {
            // Log.v("eXport-it SyncReceivingAction","Error executing local action: " + ex);

            invocation = new ActionInvocation(ex);
            responseMessage = new OutgoingActionResponseMessage(UpnpResponse.Status.INTERNAL_SERVER_ERROR);

        } catch (UnsupportedDataException ex) {
            // Log.v("eXport-it SyncReceivingAction","Error reading action request XML body: " + ex.toString(), Exceptions.unwrap(ex));

            invocation =
                    new ActionInvocation(
                            Exceptions.unwrap(ex) instanceof ActionException
                                    ? (ActionException)Exceptions.unwrap(ex)
                                    : new ActionException(ErrorCode.ACTION_FAILED, ex.getMessage())
                    );
            responseMessage = new OutgoingActionResponseMessage(UpnpResponse.Status.INTERNAL_SERVER_ERROR);

        } finally {

            if (responseMessage != null && extraResponseHeadersThreadLocal.get() != null) {
                // Log.v("eXport-it SyncReceivingAction","Merging extra headers into action response message: " + extraResponseHeadersThreadLocal.get().size());
                responseMessage.getHeaders().putAll(extraResponseHeadersThreadLocal.get());
            }

            // Always clean the TL
            requestThreadLocal.set(null);
            extraResponseHeadersThreadLocal.set(null);
        }

        try {

            // Log.v("eXport-it SyncReceivingAction","Writing body of response message");
            getUpnpService().getConfiguration().getSoapActionProcessor().writeBody(responseMessage, invocation);

            // Log.v("eXport-it SyncReceivingAction","Returning finished response message: " + responseMessage);
            return responseMessage;

        } catch (UnsupportedDataException ex) {
            // Log.v("eXport-it SyncReceivingAction","Failure writing body of response message, sending '500 Internal Server Error' without body");
            // Log.v("eXport-it SyncReceivingAction","Exception root cause: ", Exceptions.unwrap(ex));
            return new StreamResponseMessage(UpnpResponse.Status.INTERNAL_SERVER_ERROR);
        }
    }

    public static IncomingActionRequestMessage getRequestMessage() {
        return requestThreadLocal.get();
    }

    public static UpnpHeaders getExtraResponseHeaders() {
        return extraResponseHeadersThreadLocal.get();
    }
}
