/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.support.contentdirectory.callback;

import org.teleal.cling.controlpoint.ActionCallback;
import org.teleal.cling.model.action.ActionException;
import org.teleal.cling.model.action.ActionInvocation;
import org.teleal.cling.model.meta.Service;
import org.teleal.cling.model.types.ErrorCode;
import org.teleal.cling.model.types.UnsignedIntegerFourBytes;
import org.teleal.cling.support.contentdirectory.DIDLParser;
import org.teleal.cling.support.model.BrowseFlag;
import org.teleal.cling.support.model.BrowseResult;
import org.teleal.cling.support.model.DIDLContent;
import org.teleal.cling.support.model.SortCriterion;

// import java.util.logging.Logger;
// import android.util.Log;

/**
 * Invokes a "Browse" action, parses the result.
 *
 * @author Christian Bauer
 */
public abstract class Browse extends ActionCallback {

    public static final String CAPS_WILDCARD = "*";
    private final static String LOGTAG = "eXport-it Browse";
    
    public enum Status {
        NO_CONTENT("No Content"),
        LOADING("Loading..."),
        OK("OK");

        private String defaultMessage;

        Status(String defaultMessage) {
            this.defaultMessage = defaultMessage;
        }

        public String getDefaultMessage() {
            return defaultMessage;
        }
    }

    // private static Logger log = Logger.getLogger(Browse.class.getName());

    /**
     * Browse with first result 0 and {@link #getDefaultMaxResults()}, filters with {@link #CAPS_WILDCARD}.
     */
    public Browse(Service service, String containerId, BrowseFlag flag) {
        this(service, containerId, flag, CAPS_WILDCARD, 0, null);
    }

    /**
     * @param maxResults Can be <code>null</code>, then {@link #getDefaultMaxResults()} is used.
     */
    public Browse(Service service, String objectID, BrowseFlag flag,
                                String filter, long firstResult, Long maxResults, SortCriterion... orderBy) {

        super(new ActionInvocation(service.getAction("Browse")));

        // Log.v("eXport-it Browse", "Creating browse action for object ID: " + objectID);

        getActionInvocation().setInput("ObjectID", objectID);
        getActionInvocation().setInput("BrowseFlag", flag.toString());
        getActionInvocation().setInput("Filter", filter);
        getActionInvocation().setInput("StartingIndex", new UnsignedIntegerFourBytes(firstResult));
        getActionInvocation().setInput("RequestedCount",
                new UnsignedIntegerFourBytes(maxResults == null ? getDefaultMaxResults() : maxResults)
        );
        getActionInvocation().setInput("SortCriteria", SortCriterion.toString(orderBy));
    }

    @Override
    public void run() {
        updateStatus(Status.LOADING);
        super.run();
    }

    public void success(ActionInvocation invocation) {
        // Log.v("eXport-it Browse","Successful browse action, reading output argument values");
        BrowseResult result = null;
        String res = "";
        UnsignedIntegerFourBytes nb = new UnsignedIntegerFourBytes(0L);
        UnsignedIntegerFourBytes matches = new UnsignedIntegerFourBytes(0L);
        UnsignedIntegerFourBytes updid = new UnsignedIntegerFourBytes(0L);
        if (invocation != null) {
            if (invocation.getOutput("Result") != null) res = invocation.getOutput("Result").getValue().toString();
            if (invocation.getOutput("NumberReturned") != null) nb = (UnsignedIntegerFourBytes) invocation.getOutput("NumberReturned").getValue();
            if (invocation.getOutput("TotalMatches") != null) matches = (UnsignedIntegerFourBytes) invocation.getOutput("TotalMatches").getValue();
            if (invocation.getOutput("UpdateID") != null) updid = (UnsignedIntegerFourBytes) invocation.getOutput("UpdateID").getValue();
        }

        try {
          result = new BrowseResult(res, nb, matches, updid);
        } catch(NullPointerException e) {
        } catch (Exception ex) {
        }
        
        boolean proceed = receivedRaw(invocation, result);

        
        if ((result != null) && (proceed && result.getCountLong() > 0 && result.getResult().length() > 0)) {

            try {

                DIDLParser didlParser = new DIDLParser();
                DIDLContent didl = didlParser.parse(result.getResult());
                received(invocation, didl);
                updateStatus(Status.OK);

            } catch (Exception ex) {
                invocation.setFailure(
                        new ActionException(ErrorCode.ACTION_FAILED, "Can't parse DIDL XML response: " + ex, ex)
                );
                // Log.v("eXport-it Browse","Browse DIDLParser exception: " + ex);
                failure(invocation, null);
            }

        } else {
            // Log.v("eXport-it Browse","Browse received NO content... ");
            received(invocation, new DIDLContent());
            updateStatus(Status.NO_CONTENT);
        }
    }

    /**
     * Some media servers will crash if there is no limit on the maximum number of results.
     *
     * @return The default limit, 999.
     */
    public long getDefaultMaxResults() {
        return 9999;
    }

    public boolean receivedRaw(ActionInvocation actionInvocation, BrowseResult browseResult) {
        /*
        if (log.isLoggable(Level.FINER)) {
            log.finer("-------------------------------------------------------------------------------------");
            log.finer("\n" + XML.pretty(browseResult.getDidl()));
            log.finer("-------------------------------------------------------------------------------------");
        }
        */
 
        int l = 0;
        int i = 0;
        String line = "";
        String logline = "";
        // Log.v(LOGTAG, " result length: "+browseResult.getResult().length());
        l = browseResult.getResult().length();
        line = browseResult.getResult();
        while (l > 0) {
           if (l >= 1000) {   
             logline = line.substring(i,i+1000);
             i = i + 1000;
             l = l -1000;
           } else {
             logline = line.substring(i); 
             i = i + l;
             l = 0;
           }  
           // Log.v(LOGTAG, " result: "+logline); 
        }  
        return true;
    }

    public abstract void received(ActionInvocation actionInvocation, DIDLContent didl);
    public abstract void updateStatus(Status status);

}
