/*
 * Copyright (C) 2011 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.transport;

import org.teleal.cling.UpnpServiceConfiguration;
import org.teleal.cling.model.NetworkAddress;
import org.teleal.cling.model.message.IncomingDatagramMessage;
import org.teleal.cling.model.message.OutgoingDatagramMessage;
import org.teleal.cling.model.message.StreamRequestMessage;
import org.teleal.cling.model.message.StreamResponseMessage;
import org.teleal.cling.protocol.ProtocolFactory;
import org.teleal.cling.transport.spi.InitializationException;
import org.teleal.cling.transport.spi.NetworkAddressFactory;
import org.teleal.cling.common.util.Exceptions;
import org.teleal.cling.transport.spi.DatagramIO;
import org.teleal.cling.transport.spi.MulticastReceiver;
import org.teleal.cling.transport.spi.StreamClient;
import org.teleal.cling.transport.spi.StreamServer;

import java.net.InetAddress;
import java.net.NetworkInterface;
import java.util.Collections;
import java.util.List;

// import android.util.Log;

/**
 * Default implementation of switchable network message router.
 * <p>
 * This implementation is actually wrapping a regular
 * {@link org.teleal.cling.transport.RouterImpl}, making it on/off
 * switchable.
 * </p>
 * <p>
 * If the router can't be enabled (e.g. an exception occurs during
 * socket binding), a warning log message will be printed. You can customize this behavior by
 * overriding {@link #handleStartFailure(org.teleal.cling.transport.spi.InitializationException)}.
 * </p>
 * <p>
 * You have to expect that {@link #disable()} and @{link #enable()} will throw a
 * {@link RouterLockAcquisitionException}. We do not manually abort ongoing HTTP stream
 * connections, they might block access to the underlying synchronized router instance
 * when you try to disable/enable it. However, the only situation in which you should get this
 * exception is when an the response from an HTTP server is taking longer than the timeout
 * of this classes' lock acquisition routine, see {@link #getLockTimeoutMillis()}. This
 * might be the case if the response is large and/or the network connection is slow, but
 * not slow enough to run into connection/data read timeouts.
 * </p>
 *
 * @author Christian Bauer
 */
public abstract class SwitchableRouterImpl implements SwitchableRouter {

    // final private static Logger log = Logger.getLogger(Router.class.getName());

    final private UpnpServiceConfiguration configuration;
    final private ProtocolFactory protocolFactory;
    private InetAddress addr;

    private Router router;

    public SwitchableRouterImpl(UpnpServiceConfiguration configuration, ProtocolFactory protocolFactory, InetAddress addr) {
        this.configuration = configuration;
        this.protocolFactory = protocolFactory;
        this.addr = addr;
    }

    public UpnpServiceConfiguration getConfiguration() {
        return configuration;
    }

    public ProtocolFactory getProtocolFactory() {
        return protocolFactory;
    }

    public boolean isEnabled() {
        return router != null;
    }

    public boolean enable() {
            if (router == null) {
                try {
                    // log.fine("Enabling network transport router");
                    router = new RouterImpl(getConfiguration(), getProtocolFactory(), addr);
                    return true;
                } catch (InitializationException ex) {
                    handleStartFailure(ex);
                }
            }
            return false;
    }

    public void handleStartFailure(InitializationException ex) {
        // log.severe("Unable to initialize network router: " + ex);
        // log.severe("Cause: " + Exceptions.unwrap(ex));
    }

    public boolean disable() {
            if (router != null) {
                // log.fine("Disabling network transport router");
                router.shutdown();
                router = null;
                return true;
            }
            return false;
    }

}
