/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Copyright (C) March 2020 Design & Development of Computerized Solutions, Belgium
 * program modified by Philippe Gillain for eXport-it application, removing Apache HTTP Client library requirement.
 *
 */

package org.teleal.cling.transport.impl;

import org.teleal.cling.model.message.OutgoingDatagramMessage;
import org.teleal.cling.transport.Router;
import org.teleal.cling.transport.spi.DatagramIO;
import org.teleal.cling.transport.spi.DatagramProcessor;
import org.teleal.cling.transport.spi.InitializationException;
import org.teleal.cling.transport.spi.UnsupportedDataException;
import org.teleal.cling.protocol.ProtocolCreationException;
import org.teleal.cling.protocol.ProtocolFactory;
import org.teleal.cling.protocol.ReceivingAsync;
import org.teleal.cling.UpnpServiceConfiguration;
import org.teleal.cling.common.util.Exceptions;
import org.teleal.cling.model.message.IncomingDatagramMessage;
import org.teleal.cling.protocol.ProtocolCreationException;
import org.teleal.cling.protocol.ReceivingAsync;

import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
// import java.util.logging.Logger;

import java.io.IOException;
import java.net.Inet6Address;
import java.net.NetworkInterface;
import java.net.SocketAddress;
import java.net.UnknownHostException;
import java.util.Enumeration;
// import java.util.logging.Level;

/**
 * Default implementation based on a single shared (receive/send) UDP <code>MulticastSocket</code>.
 * <p>
 * Although we do not receive multicast datagrams with this service, sending multicast
 * datagrams with a configuration time-to-live requires a <code>MulticastSocket</code>.
 * </p>
 * <p>
 * Thread-safety is guaranteed through synchronization of methods of this service and
 * by the thread-safe underlying socket.
 * </p>
 * @author Christian Bauer
 */
public class DatagramIOImpl implements DatagramIO<DatagramIOConfigurationImpl> {

    
    final protected DatagramIOConfigurationImpl configuration;

    protected Router router;
    public DatagramProcessor datagramProcessor;
    protected final UpnpServiceConfiguration upnpConfiguration;
    protected final ProtocolFactory protocolFactory;
    protected InetAddress bindAddress = null;
    protected InetAddress bindAddress6 = null;
    protected MulticastSocket socket; // For sending unicast & multicast, and receiving unicast
    protected InetSocketAddress multicastAddress;
    NetworkInterface selectedIntf = null;
    
    public DatagramIOImpl(DatagramIOConfigurationImpl configuration, UpnpServiceConfiguration upnpConfiguration,  ProtocolFactory protocolFactory) {
        this.configuration = configuration;
        this.protocolFactory = protocolFactory;
        this.upnpConfiguration = upnpConfiguration;
        this.bindAddress = this.upnpConfiguration.getLocalInetAddress();
        this.bindAddress6 = this.upnpConfiguration.getLocalInetAddress6();
        router = this.upnpConfiguration.getRouter();
        datagramProcessor = this.upnpConfiguration.getDatagramProcessor();

    }
    
    @Override
    public DatagramIOConfigurationImpl getConfiguration() {
        return configuration;
    }
    
    @Override
    synchronized public void init(InetAddress bindAddress, Router router, final DatagramProcessor datagramProcessor) {

        this.router = router;
        
    }
    
    @Override
    synchronized public void stop() {
        if (socket != null && !socket.isClosed()) {
           try { 
               socket.close();
           } catch (Exception ex) { } 
        }
    }
    
    @Override
    public void run() {
 
        // Log.v("eXport-it DatagramIO","calling init procedure... ");
        init(bindAddress, router, datagramProcessor);
        try {
          // socket = new MulticastSocket(1901);
          socket = new MulticastSocket();
          // Log.v("eXport-it DatagramIO","socket created!!!!");
          if (socket != null) {
            socket.setReuseAddress(true);
            socket.setTimeToLive(configuration.getTimeToLive()); 
            // socket.setLoopbackMode(false);
            socket.setReceiveBufferSize(32768); // Keep a backlog of incoming datagrams if we are not fast enough

                for (Enumeration<NetworkInterface> en = NetworkInterface.getNetworkInterfaces(); en.hasMoreElements(); ) {
                    NetworkInterface intf = en.nextElement();
                    String intfname = intf.getName();
                    // Log.v("eXport-it DatagramIO","interface "+intfname+"   multicast: "+intf.supportsMulticast()+" up: "+intf.isUp());
                    if (intf.isUp() && intf.supportsMulticast()) {
                       for (Enumeration<InetAddress> enumIpAddr = intf.getInetAddresses(); enumIpAddr.hasMoreElements();) {
                          InetAddress inetAddress = enumIpAddr.nextElement();
                          if (inetAddress != null) {
                             // if (bindAddress!= null) {
                                // if (bindAddress.getHostAddress() == inetAddress.getHostAddress()) {
                                   selectedIntf = intf;  
                                   // Log.v("eXport-it DatagramIO","setNetworkInterface: "+ intf.getName()+"  ip:"+inetAddress.getHostAddress());
                                   socket.setNetworkInterface(selectedIntf);
                                // }
                             // } else {
                                bindAddress = inetAddress; 
                             // }
                             // Log.v("eXport-it DatagramIO","joinGroup on: "+ intf.getName()+"  ip:"+inetAddress.getHostAddress());
                             socket.joinGroup(multicastAddress,selectedIntf);
                          }   
                       }   
                    }
                }    
          }           

        } catch (IOException ioe) { 
            // Log.v("eXport-it DatagramIO","setting socket IOexception  "+ioe.toString());
        } catch (Exception ex) {
            // Log.v("eXport-it DatagramIO","setting socket exception  "+ex.toString());
            // throw new InitializationException("Could not initialize " + getClass().getSimpleName() + ": " + ex);
        } 
        
      if (socket != null) {   
        // Log.v("eXport-it DatagramIO","Entering blocking receiving loop, listening for UDP datagrams on: " + socket.getLocalAddress());
        while (true) {

            try {
                //byte[] buf = new byte[getConfiguration().getMaxDatagramBytes()];
                byte[] buf = new byte[4096];
                DatagramPacket datagram = new DatagramPacket(buf, buf.length);

                socket.receive(datagram);
                /* Log.v("eXport-it DatagramIO",
                        "UDP datagram received from: "
                                + datagram.getAddress().getHostAddress()
                                + ":" + datagram.getPort()
                                + " on: " + bindAddress.getHostAddress()
                ); */
                InetAddress senderAddress =  datagram.getAddress();
                if ((senderAddress instanceof Inet6Address) && (bindAddress6 != null)) {
                  received(datagramProcessor.read(bindAddress6, datagram));  
                } else {
                  received(datagramProcessor.read(bindAddress, datagram));
                }
            // } catch (SocketTimeoutException e) {
            //    continue;
            } catch (SocketException ex) {
                // Log.v("eXport-it DatagramIO","Socket Exception " + ex.getMessage());
                break;
            } catch (IOException ioe) {    
            } catch (UnsupportedDataException ex) {
                // Log.v("eXport-it DatagramIO","Could not read datagram: " + ex.getMessage());
            } catch (Exception ex) {
                throw new RuntimeException(ex);
            }
        }
        try {
            if (!socket.isClosed()) {
                // Log.v("eXport-it DatagramIO","Closing unicast socket");
                socket.close();
            }
        } catch (Exception ex) {
            // throw new RuntimeException(ex);
        }
      }  
    }
    
    @Override
    synchronized public void sendAsync(OutgoingDatagramMessage message) {
      // Log.v("eXport-it DatagramIO","Sending message from address: " + bindAddress.getHostAddress());
      if ((datagramProcessor != null) && (message != null)) {
        DatagramPacket packet = null;  
        try {  
           packet = datagramProcessor.write(message);
        } catch (UnsupportedDataException uex) { 
            // Log.v("eXport-it DatagramIO","Sending Async DataException: " + uex.toString());
        } catch (Exception ex) { 
            // Log.v("eXport-it DatagramIO","Sending Async Exception: " + ex.toString());
        } 
        // Log.v("eXport-it DatagramIO","Sending Async UDP datagram packet to: " + message.getDestinationAddress() + ":" + message.getDestinationPort());
        if (packet != null) {
            send(packet);
        }
      }  
    }
    @Override
    synchronized public void send(DatagramPacket datagram) {
      // Log.v("eXport-it DatagramIO","Sending message from: " + bindAddress.getHostAddress()+"   to: "+datagram.getAddress().toString()+":"+datagram.getPort());
      if ((socket != null) && (datagram !=  null)) {
        try {
            socket.send(datagram);
        } catch (SocketException ex) {
            // Log.v("eXport-it DatagramIO","Send, aborting datagram send to: " + datagram.getAddress()+"  SocketException " + ex.toString());
        } catch (IOException ioe) { 
            // Log.v("eXport-it DatagramIO","Send , aborting datagram send to: " + datagram.getAddress()+"  IOException " + ioe.toString());
        } catch (RuntimeException ex) {
            // Log.v("eXport-it DatagramIO","Send, aborting datagram send to: " + datagram.getAddress()+"  RuntimeException " + ex.toString());
            // throw ex;
        } catch (Exception ex) {
            // Log.v("eXport-it DatagramIO","Send , aborting datagram send to: " + datagram.getAddress()+"  Exception " + ex.toString());
            // throw new RuntimeException(ex);
        }
      }  
    }
    
     /**
     * Obtains the asynchronous protocol {@code Executor} and runs the protocol created
     * by the {@link org.teleal.cling.protocol.ProtocolFactory} for the given message.
     * <p>
     * If the factory doesn't create a protocol, the message is dropped immediately without
     * creating another thread or consuming further resoures. This means we can filter the
     * datagrams in the protocol factory and e.g. completely disable discovery or only
     * allow notification message from some known services we'd like to work with.
     * </p>
     *
     * @param msg The received datagram message.
     */
    @Override
    synchronized public void received(IncomingDatagramMessage msg) {
      if (msg != null) { 
        try {
            ReceivingAsync protocol = protocolFactory.createReceivingAsync(msg);
            if (protocol == null) {
                // Log.v("eXport-it DatagramIO","No protocol, ignoring received message: " + msg);
                return;
            }
            // Log.v("eXport-it DatagramIOImpl","Received asynchronous message: " + msg);
            upnpConfiguration.getAsyncProtocolExecutor().execute(protocol);
        } catch (ProtocolCreationException ex) {
            // Log.v("eXport-it DatagramIO","Handling received datagram failed - " + Exceptions.unwrap(ex).toString());
        }
      }  
    }   
    
}
