/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Copyright (C) March 2020 Design & Development of Computerized Solutions, Belgium
 * program modified by Philippe Gillain for eXport-it application, removing Apache HTTP Client library requirement.
 *
 */

package org.teleal.cling.transport.impl;

// import android.util.Log;

import org.teleal.cling.transport.Router;
import org.teleal.cling.transport.spi.DatagramProcessor;
import org.teleal.cling.transport.spi.InitializationException;
import org.teleal.cling.transport.spi.MulticastReceiver;
import org.teleal.cling.transport.spi.UnsupportedDataException;
import org.teleal.cling.protocol.ProtocolCreationException;
import org.teleal.cling.protocol.ProtocolFactory;
import org.teleal.cling.protocol.ReceivingAsync;
import org.teleal.cling.common.util.Exceptions;
import org.teleal.cling.model.message.IncomingDatagramMessage;
import org.teleal.cling.model.message.OutgoingDatagramMessage;
import org.teleal.cling.UpnpServiceConfiguration;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.net.NetworkInterface;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.Enumeration;
import java.util.List;
import java.util.Date;
import java.text.SimpleDateFormat;

/**
 * Default implementation based on a UDP <code>MulticastSocket</code>.
 * <p>
 * Thread-safety is guaranteed through synchronization of methods of this service and
 * by the thread-safe underlying socket.
 * </p>
 * @author Christian Bauer
 */
public class MulticastReceiverImpl implements MulticastReceiver<MulticastReceiverConfigurationImpl> {


    final protected MulticastReceiverConfigurationImpl configuration;

    protected Router router;
    public final DatagramProcessor datagramProcessor;
    protected final ProtocolFactory protocolFactory;
    protected final UpnpServiceConfiguration upnpConfiguration;
    protected InetAddress bindAddress = null;
    protected InetAddress bindAddress6 = null;
    protected InetSocketAddress multicastAddress = null;
    protected InetSocketAddress multicastAddress6 = null;    
    private MulticastSocket socket;
    // private NetworkInterface multicastInterface;
    // private WifiManager wifiManager;
    // private WifiManager.MulticastLock multicastLock;
    private DatagramPacket dg; 
    private int socketbuffersize = 32768;
    private int ttl = 8;
    private NetworkInterface selectedIntf = null;
    
    public MulticastReceiverImpl(MulticastReceiverConfigurationImpl configuration, UpnpServiceConfiguration upnpConfiguration, ProtocolFactory protocolFactory) {
        
        this.configuration = configuration;
        this.protocolFactory = protocolFactory;
        this.upnpConfiguration = upnpConfiguration;
        this.bindAddress = this.upnpConfiguration.getLocalInetAddress();
        this.bindAddress6 = this.upnpConfiguration.getLocalInetAddress6();
        // this.wifiManager = this.upnpConfiguration.getWifiManager();
        // this.multicastInterface = this.upnpConfiguration.getWifiNetworkInterface(wifiManager);
        this.socketbuffersize = this.upnpConfiguration.getSocketBufferSize();
        // this.ttl = this.upnpConfiguration.getTTL();
        datagramProcessor = this.upnpConfiguration.getDatagramProcessor();
        init(datagramProcessor);
        
    }
    
    @Override
    public MulticastReceiverConfigurationImpl getConfiguration() {
        return configuration;
    }
    
    @Override
    synchronized public void init(final DatagramProcessor datagramProcessor) throws InitializationException {

       /* if (wifiManager != null) {
              multicastLock = wifiManager.createMulticastLock("eXport-it");
              if (multicastLock != null) {
                multicastLock.acquire();
              }
        }  */
        
    }
   
    @Override
    public void run () {

        try {

            // Log.v("eXport-it MulticastRcv","Creating wildcard socket (for receiving multicast datagrams) on port: " + configuration.getPort());
            multicastAddress = new InetSocketAddress(configuration.getGroup(), configuration.getPort());
            multicastAddress6 = new InetSocketAddress(configuration.getGroup6(), configuration.getPort());
            socket = new MulticastSocket(1900);
            // Log.v("eXport-it MulticastRcv","local inet addr: "+bindAddress.getHostAddress());
            /* SocketAddress addr = new InetSocketAddress(bindAddress,configuration.getPort());
            try {
                socket.bind(addr);
            } catch (IOException ex) { } */
            if (socket != null) {
              socket.setReuseAddress(true);
              // socket.bind(new InetSocketAddress(configuration.getPort()));
              socket.setTimeToLive(ttl);
              socket.setBroadcast(false);
              // socket.setLoopbackMode(false);
              socket.setReceiveBufferSize(32768); // Keep a backlog of incoming datagrams if we are not fast enough
              // socket.setSoTimeout(20000);
              if (multicastAddress != null) {
                 for (Enumeration<NetworkInterface> en = NetworkInterface.getNetworkInterfaces(); en.hasMoreElements(); ) {
                    NetworkInterface intf = en.nextElement();
                    String intfname = intf.getName();
                    if (intf.isUp() && intf.supportsMulticast()) {
                       for (Enumeration<InetAddress> enumIpAddr = intf.getInetAddresses(); enumIpAddr.hasMoreElements();) {
                          InetAddress inetAddress = enumIpAddr.nextElement();
                          if (inetAddress != null) {
                             // if (bindAddress!= null) {
                                // if (bindAddress.getHostAddress() == inetAddress.getHostAddress()) {
                                   selectedIntf = intf;  
                                   // Log.v("eXport-it MulticastRcv","setNetworkInterface: "+ intf.getName()+"  ip:"+inetAddress.getHostAddress());
                                   socket.setNetworkInterface(selectedIntf);
                               // }
                             // } else {
                                bindAddress = inetAddress; 
                             // }
                             // Log.v("eXport-it MulticastRcv","joinGroup on: "+ intf.getName()+"  ip:"+inetAddress.getHostAddress());
                             // socket.setNetworkInterface(selectedIntf);
                             socket.joinGroup(multicastAddress,selectedIntf);
                          }   
                       }   
                    }
                 }    
              }              
              if (selectedIntf == null) { 
                if (configuration.getGroup() != null) {  
                  socket.joinGroup(configuration.getGroup());
                }  
              }   
              if (configuration.getGroup6() != null) {
                   socket.joinGroup(configuration.getGroup6()); 
              }
            }
        } catch (IOException ioe) {  
            // Log.v("eXport-it MulticastRcv","MulticastReceiverSocket IOexception: "+ ioe.toString());
        } catch (Exception ex) {
            // Log.v("eXport-it MulticastRcv","MulticastReceiverSocket exception: "+ ex.toString());
            // throw new InitializationException("Could not initialize " + getClass().getSimpleName() + ": " + ex);
        }
        
        while (true) {
          // Log.v("eXport-it MulticastRcv","Entering blocking receiving loop, listening for UDP datagrams on: " + socket.getLocalAddress());     
          InetAddress receivedOnLocalAddress = bindAddress;
          InetAddress receivedOnLocalAddress6 = bindAddress6;
          if (socket != null) {  
            try {
                byte[] buf = new byte[getConfiguration().getMaxDatagramBytes()];
                DatagramPacket datagram = new DatagramPacket(buf, buf.length);
                socket.receive(datagram);
                // SimpleDateFormat sdfTime = new SimpleDateFormat("HH:mm:ss");
                // Date now = new Date();
                // String strTime = sdfTime.format(now);
                /* Log.v("eXport-it MulticastRcv",strTime+
                     " UDP datagram received from: " + datagram.getAddress().getHostAddress() 
                     + ":" + datagram.getPort()
                     + " and on address: " + receivedOnLocalAddress.getHostAddress()
                ); */
                InetAddress senderAddress =  datagram.getAddress();
                if ((senderAddress instanceof Inet6Address) && (bindAddress6 != null)) {
                  received(datagramProcessor.read(receivedOnLocalAddress6, datagram));  
                } else {
                  received(datagramProcessor.read(receivedOnLocalAddress, datagram));
                }
                // } catch (SocketTimeoutException ex) {
                //  Log.v("eXport-it MulticastRcv","Could not receive datagram socket exception: " + ex.getMessage());
                // break;   
   
            } catch (SocketException ex) {
                // Log.v("eXport-it MulticastRcv","Could not receive datagram socket exception: " + ex.getMessage());
                break;
            } catch (IOException ioe) {   
                // Log.v("eXport-it MulticastRcv","Could not receive datagram IOException: " + ioe.getMessage());
            } catch (UnsupportedDataException ex) {
                // Log.v("eXport-it MulticastRcv","Could not read datagram: " + ex.getMessage());
            } catch (Exception ex) {
                // Log.v("eXport-it MulticastRcv"," exception: " + ex.getMessage());
                // throw new RuntimeException(ex);
            }
          }  
        }                 
    }   
    
    @Override
    synchronized public void stop() {
        if (socket != null && !socket.isClosed()) {
            try {
                // Log.v("eXport-it MulticastRcv","Leaving multicast group");
                socket.leaveGroup(configuration.getGroup());
                if (multicastAddress != null) {
                  for (Enumeration<NetworkInterface> en = NetworkInterface.getNetworkInterfaces(); en.hasMoreElements(); ) {
                    NetworkInterface intf = en.nextElement();
                    String intfname = intf.getName();
                    // if (!intf.getName().startsWith("lo")) {
                      for (Enumeration<InetAddress> enumIpAddr = intf.getInetAddresses(); enumIpAddr.hasMoreElements();) {
                        InetAddress inetAddress = enumIpAddr.nextElement();
                        if (inetAddress != null) {
                          socket.leaveGroup(multicastAddress,intf);
                        }   
                      }   
                   // }
                  }
                }     
                // Well this doesn't work and I have no idea why I get "java.net.SocketException: Can't assign requested address"
                // So... just close it and ignore the log messages
                socket.close();
            } catch (IOException ioe) {     
            } catch (Exception ex) {
                // Log.v("eXport-it MulticastRcv","Could not leave multicast group: " + ex);
            }

            /* if (multicastLock != null && multicastLock.isHeld()) {
              multicastLock.release();
              multicastLock = null;
            } */
        }
    }
    
    /**
     * Obtains the asynchronous protocol {@code Executor} and runs the protocol created
     * by the {@link org.teleal.cling.protocol.ProtocolFactory} for the given message.
     * <p>
     * If the factory doesn't create a protocol, the message is dropped immediately without
     * creating another thread or consuming further resoures. This means we can filter the
     * datagrams in the protocol factory and e.g. completely disable discovery or only
     * allow notification message from some known services we'd like to work with.
     * </p>
     *
     * @param msg The received datagram message.
     */
    
    @Override
    public void received(IncomingDatagramMessage msg) {
        try {
            // Log.v("eXport-it MulticastRcv","create ReceiveAsync for datagram msg: "+msg); 
            ReceivingAsync protocol = protocolFactory.createReceivingAsync(msg);
            if (protocol == null) {
                // Log.v("eXport-it MulticastRcv","No protocol, ignoring received message: " + msg);
                return;
            }
            // Log.v("eXport-it MulticastRcv","Received asynchronous message: " + msg);
            upnpConfiguration.getAsyncProtocolExecutor().execute(protocol);
        } catch (ProtocolCreationException ex) {
            // Log.v("eXport-it MulticastRcv","Handling received datagram failed - " + Exceptions.unwrap(ex).toString());
        } catch (OutOfMemoryError e) {
            
        }
    }
    
    @Override
    public void send(DatagramPacket datagram) {
      dg =  datagram; 
      new Thread(new Runnable() {
        @Override
        public void run() {
          if (socket != null) {  
            try {
              socket.send(dg);
              // Log.v("eXport-it MulticastReceiverImpl", "Sending response Datagram");
            } catch (SocketException ex) {
              // Log.v("eXport-it MulticastReceiverImpl", "Socket closed, aborting datagram send to: " + dg.getAddress());
                        // So... just close it and ignore the log messages
            } catch (IOException ioe) {  
            } catch (RuntimeException ex) {
              // throw ex;  
            } catch (Exception ex) {
              // throw new RuntimeException(ex);
            }
          }
        }  
      }).start();           
    }

}
