/*
 * Copyright (C) 2011 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Copyright (C) March 2020 Design & Development of Computerized Solutions, Belgium
 * program modified by Philippe Gillain for eXport-it application, removing Apache HTTP Client library requirement.
 *
 */

package org.teleal.cling.transport.impl;

// import android.util.Log;

import org.teleal.cling.model.message.StreamRequestMessage;
import org.teleal.cling.model.message.StreamResponseMessage;
import org.teleal.cling.model.message.UpnpHeaders;
import org.teleal.cling.model.message.UpnpMessage;
import org.teleal.cling.model.message.UpnpRequest;
import org.teleal.cling.model.message.UpnpResponse;
import org.teleal.cling.model.message.header.ContentTypeHeader;
import org.teleal.cling.model.message.header.HostHeader;  
import org.teleal.cling.model.message.header.SoapActionHeader;
import org.teleal.cling.model.message.header.ApplicationURLHeader;
import org.teleal.cling.model.message.header.UserAgentHeader;
import org.teleal.cling.transport.spi.InitializationException;
import org.teleal.cling.transport.spi.StreamClient;
import org.teleal.cling.common.util.Exceptions;
import org.teleal.cling.common.util.ByteArray;
import org.teleal.cling.common.util.Text;
import org.teleal.cling.common.util.URIUtil;
import org.teleal.cling.common.http.Headers;
import org.teleal.cling.model.Constants;
import org.teleal.cling.transport.Router;
import org.teleal.cling.model.message.UpnpOperation;
import org.teleal.cling.model.message.header.EXTHeader;
import org.teleal.cling.model.message.header.ServerHeader;
import org.teleal.cling.model.message.header.UpnpHeader;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketTimeoutException;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URI;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;
import java.util.concurrent.Executor;
import java.util.Date;
import java.text.SimpleDateFormat;

/**
 * <p>
 * This implementation works on Android.
 * </p>

 */
public class StreamClientImpl implements StreamClient {

    private StreamClientConfigurationImpl configuration;
    private int maxTotalConnections = 1024;
    private int connectionTimeoutSeconds = 20;   // 5
    private int dataReadTimeoutSeconds = 30;     // 5
    private int retryCount = 2;
    private int socketbuffersize = 65536;
    private String productName = Constants.PRODUCT_TOKEN_NAME;
    private String productVersion = Constants.PRODUCT_TOKEN_VERSION;    
    private String contentCharset = "UTF-8"; // UDA spec says it's always UTF-8 entity content
    private boolean stopped = false;
    private static Executor syncExecutor = null;
    private static Router router;
    
    public StreamClientImpl(StreamClientConfigurationImpl configuration)  {      // throws InitializationException
        this.configuration = configuration;
        
        maxTotalConnections = configuration.getMaxTotalConnections();
        connectionTimeoutSeconds = configuration.getConnectionTimeoutSeconds();   // 5
        dataReadTimeoutSeconds = configuration.getDataReadTimeoutSeconds();     // 5
        socketbuffersize = configuration.getSocketBufferSize();
        contentCharset = configuration.getContentCharset();
        retryCount =  configuration.getRequestRetryCount();
       
        
    }

    public StreamClientConfigurationImpl getConfiguration() {
        return configuration;
    }

    @Override
    public StreamResponseMessage sendRequest(StreamRequestMessage requestMessage) {    // synchronized

        StreamResponseMessage responseMessage = null;
        String bodyresult2 = "";
        String bodyresult = "";
        int body_length = 0;
        String host = "";
        int port = 0;
        String RequestUri = "";
        String SoapAction = "";
        Socket clientSocket = null;
        InputStream in = null;
        BufferedReader br = null;
        final int bufsize = 65536;
	byte[] buf = new byte[bufsize];
	int splitbyte = 0;
	int rlen = 0;
        int read = 0;
        boolean responseHeadersReceived = false;
        boolean responseBodyReceived = false;
        int responseHeadersLength = 0;
        int responseBodyLength = 0;
        int responseContentLength = 0;
        
        if ((requestMessage == null) || (stopped)) { 
            return null;
        }
        SimpleDateFormat sdfTime = new SimpleDateFormat("HH:mm:ss");
        Date now = new Date();
        String strTime = sdfTime.format(now);
        // Log.v("eXport-it StreamClient",strTime+" SendRequest start message " + requestMessage.toString());
        // ... then grab the response
        int udaMajorVersion = requestMessage.getUdaMajorVersion();
        int udaMinorVersion = requestMessage.getUdaMinorVersion();
        UpnpRequest requestOperation = requestMessage.getOperation();
        if (requestOperation == null) { 
            // Log.v("eXport-it StreamClient","requestOperation: is NULL");
            return null;
        }
        URI requestUri = requestOperation.getURI();
        if (requestUri == null) {
            // Log.v("eXport-it StreamClient","requestURI: is NULL");
            return null;
        } else {
            RequestUri = requestUri.toString();
        }
        String httpName  = requestOperation.getHttpMethodName();
        if (httpName == null) {
            // Log.v("eXport-it StreamClient","HTTP method name: is NULL");
            return null;
        }
        UpnpHeaders upnpHeaders = requestMessage.getHeaders();
        SoapAction = "";
        if (upnpHeaders.containsKey(UpnpHeader.Type.SOAPACTION)) {
            UpnpHeader SoapUpnpHdr = upnpHeaders.getFirstHeader(UpnpHeader.Type.SOAPACTION);
            SoapActionHeader SoapActionHdr = (SoapActionHeader) SoapUpnpHdr;
            SoapAction = SoapActionHdr.getString();
            // Log.v("eXport-it StreamClient","SoapAction: "+SoapAction);
        }        
        Object body = null;
        String bodyString = "";
        byte[] bodyBytes = new byte[65536];
        if (requestMessage.hasBody()) {
            if (requestMessage.getBodyType() == UpnpMessage.BodyType.STRING) {
                bodyString = requestMessage.getBodyString();
            } else {
                bodyBytes = requestMessage.getBodyBytes();
            }
        }

        try {
            if (requestUri.getHost() != null) host = requestUri.getHost();
            if (requestUri.getPort() != -1) port = requestUri.getPort();
            if ((host == null) || (host.length() < 2) || (port < 1)) {
                // Log.v("eXport-it StreamClient","invalid destination host: "+host+"  port: "+port);
                return null;
            }
            // Log.v("eXport-it StreamClient","Request message to " + host + "   port="+port+" uri: "+RequestUri);

                InetAddress ineteAddress = InetAddress.getByName(host);
                SocketAddress socketAddress = new InetSocketAddress(ineteAddress, port);
                // create a socket
                clientSocket = new Socket();
                // this method will block no more than timeout ms.
                int timeoutInMs = connectionTimeoutSeconds*1000;   // 20 seconds
                int readTimeOut = dataReadTimeoutSeconds*1000; 
                int size =  socketbuffersize;
                clientSocket.setSoTimeout(readTimeOut);
                clientSocket.setReceiveBufferSize(size);
                clientSocket.setSendBufferSize(size);
                clientSocket.setSoLinger(false,0);
                clientSocket.setTcpNoDelay(false);
                clientSocket.connect(socketAddress, timeoutInMs);


            OutputStream theOutput = clientSocket.getOutputStream();
            // no auto-flushing
            PrintWriter bw = null;
            try {
                // bw = new PrintStream(theOutput, false,"UTF-8");
                OutputStreamWriter osw = new OutputStreamWriter(theOutput, "UTF-8");
                BufferedWriter writer = new BufferedWriter(osw, 65536);
                bw = new PrintWriter(writer,true);
            } catch (UnsupportedEncodingException uee) { 
                // Log.v("eXport-it StreamClient","UnsupportedEncodingException creating PrintStream "+uee);
                return null;
            }
            if (bw == null) {
                // Log.v("eXport-it StreamClient","DataOutputStream is NULL ");
                return null;
            }
            // native line endings are uncertain so add them manually
            bw.print(httpName+ " " + requestUri.getPath() + " HTTP/1."+udaMinorVersion+"\r\n");
            String HeadersStr = upnpHeaders.toString();
            byte[] headerBytes;
            headerBytes = HeadersStr.getBytes("UTF-8");
            java.text.SimpleDateFormat gmtFrmt = null;
            gmtFrmt = new java.text.SimpleDateFormat( "E, d MMM yyyy HH:mm:ss 'GMT'", Locale.US);
	    gmtFrmt.setTimeZone(TimeZone.getTimeZone("GMT"));
            if (gmtFrmt != null) {
                bw.print( "Date: " + gmtFrmt.format( new Date()) + "\r\n");
            }
            // Log.v("eXport-it StreamClient","method "+httpName+" request Uri "+RequestUri);
            if (httpName.toUpperCase().equals("POST")) {
                clientSocket.setSoTimeout(2*readTimeOut);
                // Log.v("eXport-it StreamClient","POST request URI: "+RequestUri);
                // Log.v("eXport-it StreamClient","POST request Headers: "+upnpHeaders.toString());
                // Log.v("eXport-it StreamClient","POST request Body: "+bodyString);
                bw.print("Accept: */*\r\n");
                bw.print("Host: "+host+":"+port+"\r\n");
                if (bodyString.length() > 1) {
                    bw.print("Accept-Charset: utf-8\r\n");
                    bw.print("Accept-Language: en-US\r\n");
                    bw.print("Content-type: text/xml;charset=UTF-8\r\n");
                    bw.print("Content-Length: " + (bodyString.length()+2) + "\r\n");
                    bw.print("Connection: Close\r\n");
                    if (SoapAction.length() > 1) {
                        bw.print("SOAPAction: " + SoapAction + "\r\n");
                    } 
                    bw.print("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.0 DLNADOC/1.50 eXport-it/1.0\r\n");
                    bw.print("\r\n");
                    bw.print(bodyString);
                    bw.print("\r\n");
                    // Log.v("eXport-it StreamClient","POST request Body length="+bodyString.length()+" "+bodyString);
                } else {
                    if (bodyBytes.length > 1) {
                        bw.print("Accept-Charset: utf-8\r\n");
                        bw.print("Accept-Language: en-US\r\n");
                        bw.print("Content-type: text/xml;charset=UTF-8\r\n");
                        try {
                            bodyString = new String(bodyBytes, Charset.defaultCharset()); 
                        } catch (Exception ex)  {  }
                        bw.print("Content-Length: " + (bodyString.length()+2)+ "\r\n");
                        bw.print("Connection: Close\r\n");
                        if (SoapAction.length() > 1) {
                            bw.print("SOAPAction: " + SoapAction + "\r\n");
                        } 
                        bw.print("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.0 DLNADOC/1.50 eXport-it/1.0\r\n");
                        bw.print("\r\n");
                        bw.print(bodyString);
                        bw.print("\r\n");
                        // Log.v("eXport-it StreamClient","POST request Body length="+bodyString.length()+" "+bodyString);
                    } else {
                        bw.print("Accept-Charset: utf-8\r\n");
                        bw.print("Accept-Language: en-US\r\n");
                        bw.print("Connection: Close\r\n");    // Keep-Alive
                        if (SoapAction.length() > 1) {
                            bw.print("SOAPAction: " + SoapAction + "\r\n");
                        } 
                        bw.print("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.1 DLNADOC/1.50 eXport-it/1.0\r\n");
                        bw.print("\r\n");
                    }
                }    
            } else {
                bw.print("Accept: text/plain,text/html,text/xml,text/*,application/xhtml+xml,application/xml,application/*;q=0.9,image/webp,*/*;q=0.8\r\n");
                // bw.print("Accept-Encoding: gzip,deflate\r\n");
                bw.print("Host: "+host+":"+port+"\r\n");
                bw.print("Accept-Charset: utf-8\r\n");
                bw.print("Accept-Language: en-US;q=0.5,en;q=0.3\r\n");
                bw.print("Connection: Close\r\n");    // keep-alive
                bw.print("DNT: 1\r\n");
                bw.print("Upgrade-Insecure-Requests: 1\r\n");
                bw.print("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.0 DLNADOC/1.50 eXport-it/1.0\r\n");
                bw.print("\r\n");
            }
            bw.flush();
        } catch (SocketException sex) {
            // Log.v("eXport-it StreamClient","SocketException processing request: "+sex); // here it blocks with isConnected failed No route to host
            return null;  
        } catch (IllegalStateException isex) { 
            // Log.v("eXport-it StreamClient","IllegalStateException processing request: "+isex);
            return null;  
        } catch (IOException ex) {
            // Log.v("eXport-it StreamClient","IOexception processing request: "+ex);
            return null;
     
        } catch (Exception ex) {
            // Log.v("eXport-it StreamClient","exception processing request: "+ex);
            return null;
        }    
    
            SimpleDateFormat sdfTime1 = new SimpleDateFormat("HH:mm:ss");
            Date now1 = new Date();
            String strTime1 = sdfTime1.format(now1);

                try {
                    Thread.sleep(200);
                } catch (InterruptedException iex) {
                        
                }
            
            String line = "";
            int lineNb = 0;
            bodyresult2 = "";
            bodyresult = "";
            responseHeadersReceived = false;
            responseBodyReceived = false;
            responseHeadersLength = 0;
            responseBodyLength = 0;
            responseContentLength = 0;
            String statusLine = "";
            String HttpVersion = "";
            String StatusCode = "000";   // 3 digits but not always
            String reason_phrase = "";
            String headerKey  = "";
            String headerValue  = "";
            String contentType = "text/xml";
            String applicationUrl = "";
            boolean connectionFound = false;
            Headers headers = new Headers();
            boolean no_response = true;
            if ((clientSocket == null) || (clientSocket.isClosed())) {
                // Log.v("eXport-it StreamClient","Socket is closed before getting response!!!");
                return null;
            }
            try {
                in = clientSocket.getInputStream();
                if (in == null) return null;
                InputStreamReader isr = new InputStreamReader(in, "UTF-8");
                br = new BufferedReader(isr, 65536);
            } catch (SocketException sex) {
                // Log.v("eXport-it StreamClient","SocketException getting InputStream reading response: "+sex);
                if (in != null) {
                    try { in.close(); } catch (IOException ioe) {  }
                }    
                return null;  
            } catch (IllegalStateException isex) { 
                // Log.v("eXport-it StreamClient","IllegalStateException getting InputStream reading response: "+isex);
                if (in != null) {
                    try { in.close(); } catch (IOException ioe) {  }
                }
                return null;   
            } catch (IOException ex) {
                // Log.v("eXport-it StreamClient","IOexception getting InputStream reading response: "+ex);
                if (in != null) {
                    try { in.close(); } catch (IOException ioe) {  }
                }
                return null;    
            } catch (Exception ex) {
                // Log.v("eXport-it StreamClient","exception getting InputStream reading response: "+ex);
                if (in != null) {
                    try { in.close(); } catch (IOException ioe) {  }
                }
               return null;
            }        
                if (br != null) {     
                    // Log.v("eXport-it StreamClient","getting Response message ");
                    line = null;
                    int line_length = 0;
                    int r = 0;
                    while ((line_length < 10) && ( r < 5)) {
                        try {
                            line = br.readLine();
                            r++;
                            if (line == null) {
                                line_length = 0;
                            } else {
                                line_length = line.length();
                            }
                        } catch (SocketException sex) {
                            // Log.v("eXport-it StreamClient","SocketException InputStream waiting for input on response: "+sex);  // Connection Reset !!!
                            if (line.length() == 0) { 
                                if (in != null) {
                                   try { in.close(); } catch (IOException ioe) {  }
                                }
                                return null;
                            } else {
                                // Log.v("eXport-it StreamClient","SocketException InputStream waiting for input received: "+line);
                            }
                        } catch (IllegalStateException isex) { 
                            // Log.v("eXport-it StreamClient","IllegalStateException InputStream waiting for input on response: "+isex);
                            if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                            }
                            return null;   
                        } catch (IOException ex) {
                            // Log.v("eXport-it StreamClient","IOexception InputStream waiting for input on response: "+ex);
                            if (in != null) {
                                try { in.close(); } catch (IOException ioe) {  }
                            }
                            return null;    
                        } catch (Exception ex) {
                            // Log.v("eXport-it StreamClient","exception InputStream waiting for input on response: "+ex);
                            if (in != null) {
                                try { in.close(); } catch (IOException ioe) {  }
                            }
                            return null;
                        }
                    }
                    while (line != null) {
                        bodyresult2 = bodyresult2 + line + "\r\n";
                        if (line.length() != 0) {
                            if (lineNb == 0) { // Status line
                                statusLine = line;    // status-line = HTTP-version SP status-code SP reason-phrase CRLF
                                String[] wrd = statusLine.split(" ");
                                if (wrd != null) {
                                    if (wrd.length > 1) {
                                        HttpVersion = wrd[0];
                                        StatusCode = wrd[1];
                                        if (wrd.length > 2) {
                                            for (int j = 2; j < wrd.length; j++) {
                                                reason_phrase = reason_phrase + wrd[j];
                                            }
                                        }
                                        // Log.v("eXport-it StreamClient","Response message "+HttpVersion+" ReturnCode= "+StatusCode);
                                        no_response = false;
                                    }
                                }
                            } else {
                                String[] wrd2 = line.split(" ");
                                applicationUrl =  "";
                                if (wrd2 != null) {
                                    if (wrd2.length > 1) {
                                        headerKey = wrd2[0];
                                        headerKey = headerKey.replace(":","");
                                        int l = wrd2.length;
                                        headerValue = wrd2[1];
                                        for (int k=2; k < l; k++) {
                                            headerValue = headerValue + " " + wrd2[k];
                                        }
                                        if (headerKey.toUpperCase().equals("CONTENT-TYPE")) {
                                            contentType =  headerValue; 
                                        } else {
                                            if (headerKey.toUpperCase().equals("APPLICATION-URL")) {
                                                applicationUrl =  headerValue; 
                                                // Log.v("eXport-it StreamClient","<application URL header found: "+applicationUrl);
                                            } else {
                                                if (headerKey.toUpperCase().equals("CONNECTION")) {
                                                    // Log.v("eXport-it StreamClient","Connection header found: "+headerValue);
                                                    connectionFound = true; 
                                                } else {
                                                    if (headerKey.toUpperCase().equals("CONTENT-LENGTH")) {
                                                        String contentLength =  headerValue; 
                                                        try { responseContentLength = Integer.parseInt(contentLength);
                                                        } catch (NumberFormatException ex) { }
                                                    }
                                                }
                                            }
                                        }
                                        headers.add(headerKey, headerValue);
                                    }
                                }
                            }
                            lineNb++;
                        } else {
                            if (lineNb == 0) {
                                // Log.v("eXport-it StreamClient","null line found, trying to read response!!!"); 
                            }
                            break; 
                        }
                        try {
                            line = br.readLine();
                        } catch (SocketException sex) {
                           break;
                        } catch (IllegalStateException isex) { 
                           // Log.v("eXport-it StreamClient","IllegalStateException reading following Header Lines of response: "+isex);
                           if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                           }
                           return null;   
                        } catch (IOException ex) {
                           // Log.v("eXport-it StreamClient","IOexception reading following Header Lines of response: "+ex);
                        } catch (Exception ex) {
                           // Log.v("eXport-it StreamClient","Exception reading following Header Lines of response: "+ex); 
                           if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                           }
                           return null;
                        }    
                    }
 
                    if ((headers != null) && (!connectionFound)) {
                         headers.add("Connection", "close");
                    }
                    line = null;
                    try {
                        line = br.readLine();
                    } catch (SocketException sex) {
                       // Log.v("eXport-it StreamClient","SocketException reading initial Body line of response: "+sex);
                    } catch (IllegalStateException isex) { 
                      // Log.v("eXport-it StreamClient","IllegalStateException reading initial Body line of response: "+isex);
                      if (in != null) {
                          try { in.close(); } catch (IOException ioe) {  }
                      }
                      return null;   
                    } catch (IOException ex) {
                      // Log.v("eXport-it StreamClient","IOexception reading initial Body line of response: "+ex);
                    } catch (Exception ex) {
                      // Log.v("eXport-it StreamClient","exception reading initial Body line of response: "+ex);
                      if (in != null) {
                          try { in.close(); } catch (IOException ioe) {  }
                      }
                      return null;
                    }  
                    while (line != null) {      // && (line.length() >= 0))
                        bodyresult = bodyresult + line + "\r\n";
                        try {
                            line = br.readLine();
                        } catch (SocketException sex) {
                           // Log.v("eXport-it StreamClient","SocketException reading next Body line of response: "+sex);
                        } catch (IllegalStateException isex) { 
                           // Log.v("eXport-it StreamClient","IllegalStateException reading next Body line of response: "+isex);
                           if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                           }
                           return null;   
                        } catch (IOException ex) {
                           // Log.v("eXport-it StreamClient","IOexception reading next Body line of response: "+ex);
                        } catch (Exception ex) {
                           // Log.v("eXport-it StreamClient","Exception reading next Body line of response: "+ex);
                           if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                           }
                           return null;
                        }
                    }
                    if (bodyresult != null) { body_length = bodyresult.length();
                    } else { body_length = 0; }

                } else {
                    if (in != null) {
                        try { in.close(); } catch (IOException ioe) {  }
                    }
                    return null;
                }
            
            if (bodyresult.length() > 2) {
                if (httpName.toUpperCase().equals("POST")) {
                    int lng = 0;
                    try {
                        byte[] BodyByte = bodyresult.getBytes("UTF-8");
                        lng = BodyByte.length;
                    } catch (UnsupportedEncodingException e) {
                    }
                    // Log.v("eXport-it StreamClient","POST response Body length="+bodyresult.length()+"/"+lng+" "+bodyresult);
                }
                responseMessage = new StreamResponseMessage(bodyresult, new ContentTypeHeader(contentType));
                if (responseMessage != null) {
                    UpnpHeaders upnpHeaders2 = new UpnpHeaders(); 
                    if ((headers != null) && (upnpHeaders2 != null)) {
                        for (Map.Entry<String, List<String>> entry : headers.entrySet()) {
                            for (String value : entry.getValue())  {
                                upnpHeaders2.add(entry.getKey(), value);
                                /* if (entry.getKey().equals(UpnpHeader.Type.APPLICATIONURL)) {
                                    Log.v("eXport-it StreamClient"," ApplicationURL header added to upnpHeader2!!! "+applicationUrl);
                                } */
                            }
                        }
                    }
                    if (!upnpHeaders2.containsKey(UpnpHeader.Type.HOST)) {
                        upnpHeaders2.add(UpnpHeader.Type.HOST, new HostHeader(host, port));
                    }
                    if (!upnpHeaders2.containsKey(UpnpHeader.Type.EXT)) {
                        upnpHeaders2.add(UpnpHeader.Type.EXT, new EXTHeader());
                    }
                    if (!upnpHeaders2.containsKey(UpnpHeader.Type.USER_AGENT)) {
                        UserAgentHeader userAgentHdr = new UserAgentHeader();
                        userAgentHdr.setString("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.0 DLNADOC/1.50 eXport-it/1.0");
                        upnpHeaders2.add(UpnpHeader.Type.USER_AGENT,userAgentHdr);
                    }
                    if (applicationUrl.length() > 1) {
                        // Log.v("eXport-it StreamClient","response message setting ApplicationURL: "+applicationUrl);
                        ApplicationURLHeader applUrlHdr = new ApplicationURLHeader();
                        applUrlHdr.setString(applicationUrl);
                        upnpHeaders2.add(UpnpHeader.Type.APPLICATIONURL, applUrlHdr); 
                    }
                    responseMessage.setHeaders(upnpHeaders2);
                } else {
                    // Log.v("eXport-it StreamClient","Stream response message is NULL!!!!");
                    return null;
                }
            } else {
                // Log.v("eXport-it StreamClient","getting Response message with empty body!!! "+bodyresult2);
                if (bodyresult2.length() < 2) {
                   return null;
                }
            }

        try {    
            br.close();
            in.close();
            // theOutput.close();
        } catch (SocketException sex) {
            // Log.v("eXport-it StreamClient","SocketException reading response: "+sex);
            return null;  
        } catch (IllegalStateException isex) { 
            // Log.v("eXport-it StreamClient","IllegalStateException reading response: "+isex);
            return null;   
        } catch (IOException ex) {
            // Log.v("eXport-it StreamClient","IOexception reading response: "+ex);
            return null;    
        } catch (Exception ex) {
            // Log.v("eXport-it StreamClient","exception reading response: "+ex);
            return null;
        }    

        if (clientSocket != null) {
            if (!clientSocket.isClosed()) {
                try {
                     clientSocket.close();
                 } catch (Exception ex) { }
            }
        }
        // Date end = new Date();
        // strTime = sdfTime.format(end);
        // Log.v("eXport-it StreamClient",strTime+" SendRequest end message " + requestMessage.toString()+" response body length: "+body_length);
        // Log.v("eXport-it StreamClient","returning Response message body length: "+bodyresult.length());
        return responseMessage;
        
    }

    @Override
    public synchronized StreamResponseMessage sendRequestSync(StreamRequestMessage requestMessage) {    // synchronized

        StreamResponseMessage responseMessage = null;
        String bodyresult2 = "";
        String bodyresult = "";
        int body_length = 0;
        String host = "";
        int port = 0;
        String RequestUri = "";
        String SoapAction = "";
        Socket clientSocket = null;
        InputStream in = null;
        BufferedReader br = null;
        final int bufsize = 65536;
	byte[] buf = new byte[bufsize];
	int splitbyte = 0;
	int rlen = 0;
        int read = 0;
        boolean responseHeadersReceived = false;
        boolean responseBodyReceived = false;
        int responseHeadersLength = 0;
        int responseBodyLength = 0;
        int responseContentLength = 0;
        
        if ((requestMessage == null) || (stopped)) { 
            return null;
        }
        SimpleDateFormat sdfTime = new SimpleDateFormat("HH:mm:ss");
        Date now = new Date();
        String strTime = sdfTime.format(now);
        // Log.v("eXport-it StreamClient",strTime+" SendRequest start message " + requestMessage.toString());
        // ... then grab the response
        int udaMajorVersion = requestMessage.getUdaMajorVersion();
        int udaMinorVersion = requestMessage.getUdaMinorVersion();
        UpnpRequest requestOperation = requestMessage.getOperation();
        if (requestOperation == null) { 
            // Log.v("eXport-itStreamClientS","requestOperation: is NULL");
            return null;
        }
        URI requestUri = requestOperation.getURI();
        if (requestUri == null) {
            // Log.v("eXport-itStreamClientS","requestURI: is NULL");
            return null;
        } else {
            RequestUri = requestUri.toString();
        }
        String httpName  = requestOperation.getHttpMethodName();
        if (httpName == null) {
            // Log.v("eXport-itStreamClientS","HTTP method name: is NULL");
            return null;
        }
        UpnpHeaders upnpHeaders = requestMessage.getHeaders();
        SoapAction = "";
        if (upnpHeaders.containsKey(UpnpHeader.Type.SOAPACTION)) {
            UpnpHeader SoapUpnpHdr = upnpHeaders.getFirstHeader(UpnpHeader.Type.SOAPACTION);
            SoapActionHeader SoapActionHdr = (SoapActionHeader) SoapUpnpHdr;
            SoapAction = SoapActionHdr.getString();
            // Log.v("eXport-it StreamClient","SoapAction: "+SoapAction);
        }        
        Object body = null;
        String bodyString = "";
        byte[] bodyBytes = new byte[65536];
        if (requestMessage.hasBody()) {
            if (requestMessage.getBodyType() == UpnpMessage.BodyType.STRING) {
                bodyString = requestMessage.getBodyString();
            } else {
                bodyBytes = requestMessage.getBodyBytes();
            }
        }

        try {
            if (requestUri.getHost() != null) host = requestUri.getHost();
            if (requestUri.getPort() != -1) port = requestUri.getPort();
            if ((host == null) || (host.length() < 2) || (port < 1)) {
                // Log.v("eXport-itStreamClientS","invalid destination host: "+host+"  port: "+port);
                return null;
            }
            // Log.v("eXport-it StreamClient","Request message to " + host + "   port="+port+" uri: "+RequestUri);

                InetAddress ineteAddress = InetAddress.getByName(host);
                SocketAddress socketAddress = new InetSocketAddress(ineteAddress, port);
                // create a socket
                clientSocket = new Socket();
                // this method will block no more than timeout ms.
                int timeoutInMs = connectionTimeoutSeconds*1000;   // 10 seconds
                int readTimeOut = dataReadTimeoutSeconds*1000; 
                int size =  socketbuffersize;
                clientSocket.setSoTimeout(readTimeOut);
                clientSocket.setReceiveBufferSize(size);
                clientSocket.setSendBufferSize(size);
                clientSocket.setSoLinger(false,0);
                clientSocket.setTcpNoDelay(false);
                clientSocket.connect(socketAddress, timeoutInMs);


            OutputStream theOutput = clientSocket.getOutputStream();
            // no auto-flushing
            PrintWriter bw = null;
            try {
                // bw = new PrintStream(theOutput, false,"UTF-8");
                OutputStreamWriter osw = new OutputStreamWriter(theOutput, "UTF-8");
                BufferedWriter writer = new BufferedWriter(osw, 65536);
                bw = new PrintWriter(writer,true);
            } catch (UnsupportedEncodingException uee) { 
                // Log.v("eXport-itStreamClientS","UnsupportedEncodingException creating PrintStream "+uee);
                return null;
            }
            if (bw == null) {
                // Log.v("eXport-itStreamClientS","DataOutputStream is NULL ");
                return null;
            }
            // native line endings are uncertain so add them manually
            bw.print(httpName+ " " + requestUri.getPath() + " HTTP/1."+udaMinorVersion+"\r\n");
            String HeadersStr = upnpHeaders.toString();
            byte[] headerBytes;
            headerBytes = HeadersStr.getBytes("UTF-8");
            java.text.SimpleDateFormat gmtFrmt = null;
            gmtFrmt = new java.text.SimpleDateFormat( "E, d MMM yyyy HH:mm:ss 'GMT'", Locale.US);
	    gmtFrmt.setTimeZone(TimeZone.getTimeZone("GMT"));
            if (gmtFrmt != null) {
                bw.print( "Date: " + gmtFrmt.format( new Date()) + "\r\n");
            }
            // Log.v("eXport-it StreamClient","method "+httpName+" request Uri "+RequestUri);
            if (httpName.toUpperCase().equals("POST")) {
                clientSocket.setSoTimeout(2*readTimeOut);
                // Log.v("eXport-itStreamClientS","POST request URI: "+RequestUri);
                // Log.v("eXport-itStreamClientS","POST request Headers: "+upnpHeaders.toString());
                // Log.v("eXport-itStreamClientS","POST request Body length: "+bodyString.length());
                bw.print("Accept: */*\r\n");
                bw.print("Host: "+host+":"+port+"\r\n");
                if (bodyString.length() > 1) {
                    bw.print("Accept-Charset: utf-8\r\n");
                    bw.print("Accept-Language: en-US\r\n");
                    bw.print("Content-type: text/xml;charset=UTF-8\r\n");
                    bw.print("Content-Length: " + (bodyString.length()+2) + "\r\n");
                    bw.print("Connection: Close\r\n");
                    if (SoapAction.length() > 1) {
                        bw.print("SOAPAction: " + SoapAction + "\r\n");
                        // Log.v("eXport-itStreamClientS","POST SoapAction: "+SoapAction);
                    } 
                    bw.print("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.0 DLNADOC/1.50 eXport-it/1.0\r\n");
                    bw.print("\r\n");
                    bw.print(bodyString);
                    bw.print("\r\n");
                    // Log.v("eXport-it StreamClient","POST request Body length="+bodyString.length()+" "+bodyString);
                } else {
                    if (bodyBytes.length > 1) {
                        bw.print("Accept-Charset: utf-8\r\n");
                        bw.print("Accept-Language: en-US\r\n");
                        bw.print("Content-type: text/xml;charset=UTF-8\r\n");
                        try {
                            bodyString = new String(bodyBytes, Charset.defaultCharset()); 
                        } catch (Exception ex)  {  }
                        bw.print("Content-Length: " + (bodyString.length()+2)+ "\r\n");
                        bw.print("Connection: Close\r\n");
                        if (SoapAction.length() > 1) {
                            bw.print("SOAPAction: " + SoapAction + "\r\n");
                            // Log.v("eXport-itStreamClientS","POST SoapAction: "+SoapAction);
                        } 
                        bw.print("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.0 DLNADOC/1.50 eXport-it/1.0\r\n");
                        bw.print("\r\n");
                        bw.print(bodyString);
                        bw.print("\r\n");
                        // Log.v("eXport-it StreamClient","POST request Body length="+bodyString.length()+" "+bodyString);
                    } else {
                        bw.print("Accept-Charset: utf-8\r\n");
                        bw.print("Accept-Language: en-US\r\n");
                        bw.print("Connection: Close\r\n");    // Keep-Alive
                        if (SoapAction.length() > 1) {
                            bw.print("SOAPAction: " + SoapAction + "\r\n");
                            // Log.v("eXport-itStreamClientS","POST SoapAction: "+SoapAction);
                        } 
                        bw.print("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.0 DLNADOC/1.50 eXport-it/1.0\r\n");
                        bw.print("\r\n");
                    }
                }    
            } else {
                bw.print("Accept: text/plain,text/html,text/xml,text/*,application/xhtml+xml,application/xml,application/*;q=0.9,image/webp,*/*;q=0.8\r\n");
                // bw.print("Accept-Encoding: gzip,deflate\r\n");
                bw.print("Host: "+host+":"+port+"\r\n");
                bw.print("Accept-Charset: utf-8\r\n");
                bw.print("Accept-Language: en-US;q=0.5,en;q=0.3\r\n");
                bw.print("Connection: Close\r\n");    // keep-alive
                bw.print("DNT: 1\r\n");
                bw.print("Upgrade-Insecure-Requests: 1\r\n");
                bw.print("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.0 DLNADOC/1.50 eXport-it/1.0\r\n");
                bw.print("\r\n");
            }
            bw.flush();
        } catch (SocketException sex) {
            // Log.v("eXport-itStreamClientS","SocketException processing request: "+sex); // here it blocks with isConnected failed No route to host
            return null;  
        } catch (IllegalStateException isex) { 
            // Log.v("eXport-itStreamClientS","IllegalStateException processing request: "+isex);
            return null;  
        } catch (IOException ex) {
            // Log.v("eXport-itStreamClientS","IOexception processing request: "+ex);
            return null;
     
        } catch (Exception ex) {
            // Log.v("eXport-itStreamClientS","exception processing request: "+ex);
            return null;
        }    
    
            SimpleDateFormat sdfTime1 = new SimpleDateFormat("HH:mm:ss");
            Date now1 = new Date();
            String strTime1 = sdfTime1.format(now1);

                try {
                    Thread.sleep(250);
                } catch (InterruptedException iex) {
                        
                }
            
            String line = "";
            int lineNb = 0;
            bodyresult2 = "";
            bodyresult = "";
            responseHeadersReceived = false;
            responseBodyReceived = false;
            responseHeadersLength = 0;
            responseBodyLength = 0;
            responseContentLength = 0;
            String statusLine = "";
            String HttpVersion = "";
            String StatusCode = "000";   // 3 digits but not always
            String reason_phrase = "";
            String headerKey  = "";
            String headerValue  = "";
            String contentType = "text/xml";
            String applicationUrl = "";
            boolean connectionFound = false;
            Headers headers = new Headers();
            boolean no_response = true;
            if ((clientSocket == null) || (clientSocket.isClosed())) {
                // Log.v("eXport-itStreamClientS","Socket is closed before getting response!!!");
                return null;
            }
            try {
                in = clientSocket.getInputStream();
                if (in == null) return null;
                InputStreamReader isr = new InputStreamReader(in, "UTF-8");
                br = new BufferedReader(isr);
            } catch (SocketException sex) {
                // Log.v("eXport-itStreamClientS","SocketException getting InputStream reading response: "+sex);
                if (in != null) {
                    try { in.close(); } catch (IOException ioe) {  }
                }    
                return null;  
            } catch (IllegalStateException isex) { 
                // Log.v("eXport-itStreamClientS","IllegalStateException getting InputStream reading response: "+isex);
                if (in != null) {
                    try { in.close(); } catch (IOException ioe) {  }
                }
                return null;   
            } catch (IOException ex) {
                // Log.v("eXport-itStreamClientS","IOexception getting InputStream reading response: "+ex);
                if (in != null) {
                    try { in.close(); } catch (IOException ioe) {  }
                }
                return null;    
            } catch (Exception ex) {
                // Log.v("eXport-itStreamClientS","exception getting InputStream reading response: "+ex);
                if (in != null) {
                    try { in.close(); } catch (IOException ioe) {  }
                }
               return null;
            }        
                if (br != null) {     
                    // Log.v("eXport-it StreamClient","getting Response message ");
                    line = null;
                    int line_length = 0;
                    int r = 0;
                    
                    while ((line_length < 10) && ( r < 5)) {
                        try {
                            line = br.readLine();
                            r++;
                            if (line == null) {
                                line_length = 0;
                            } else {
                                line_length = line.length();
                            }
                        } catch (SocketException sex) {
                            // Log.v("eXport-itStreamClientS","SocketException InputStream waiting for input on response: "+sex);  // Connection Reset !!!
                            if (line != null) {
                                if (line.length() == 0) { 
                                    if (in != null) {
                                       try { in.close(); } catch (IOException ioe) {  }
                                    }
                                    return null;
                                } else {
                                    // Log.v("eXport-itStreamClientS","SocketException InputStream waiting for input received: "+line);
                                }
                            }
                        } catch (IllegalStateException isex) { 
                            // Log.v("eXport-itStreamClientS","IllegalStateException InputStream waiting for input on response: "+isex);
                            if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                            }
                            return null;   
                        } catch (IOException ex) {
                            // Log.v("eXport-itStreamClientS","IOexception InputStream waiting for input on response: "+ex);
                            if (in != null) {
                                try { in.close(); } catch (IOException ioe) {  }
                            }
                            return null;    
                        } catch (Exception ex) {
                            // Log.v("eXport-itStreamClientS","exception InputStream waiting for input on response: "+ex);
                            if (in != null) {
                                try { in.close(); } catch (IOException ioe) {  }
                            }
                            return null;
                        }
                    }
                    while (line != null) {
                        bodyresult2 = bodyresult2 + line + "\r\n";
                        rlen = line_length;
                        if (line.length() != 0) {
                            if (lineNb == 0) { // Status line
                                statusLine = line;    // status-line = HTTP-version SP status-code SP reason-phrase CRLF
                                String[] wrd = statusLine.split(" ");
                                if (wrd != null) {
                                    if (wrd.length > 1) {
                                        HttpVersion = wrd[0];
                                        StatusCode = wrd[1];
                                        if (wrd.length > 2) {
                                            for (int j = 2; j < wrd.length; j++) {
                                                reason_phrase = reason_phrase + wrd[j];
                                            }
                                        }
                                        // Log.v("eXport-it StreamClient","Response message "+HttpVersion+" ReturnCode= "+StatusCode);
                                        no_response = false;
                                    }
                                }
                            } else {
                                String[] wrd2 = line.split(" ");
                                applicationUrl =  "";
                                if (wrd2 != null) {
                                    if (wrd2.length > 1) {
                                        headerKey = wrd2[0];
                                        headerKey = headerKey.replace(":","");
                                        int l = wrd2.length;
                                        headerValue = wrd2[1];
                                        for (int k=2; k < l; k++) {
                                            headerValue = headerValue + " " + wrd2[k];
                                        }
                                        if (headerKey.toUpperCase().equals("CONTENT-TYPE")) {
                                            contentType =  headerValue; 
                                        } else {
                                            if (headerKey.toUpperCase().equals("APPLICATION-URL")) {
                                                applicationUrl =  headerValue; 
                                                // Log.v("eXport-it StreamClient","<application URL header found: "+applicationUrl);
                                            } else {
                                                if (headerKey.toUpperCase().equals("CONNECTION")) {
                                                    // Log.v("eXport-it StreamClient","Connection header found: "+headerValue);
                                                    connectionFound = true; 
                                                } else {
                                                    if (headerKey.toUpperCase().equals("CONTENT-LENGTH")) {
                                                        String contentLength =  headerValue; 
                                                        try { responseContentLength = Integer.parseInt(contentLength);
                                                        } catch (NumberFormatException ex) { }
                                                    }
                                                }
                                            }
                                        }
                                        headers.add(headerKey, headerValue);
                                    }
                                }
                            }
                            lineNb++;
                        } else {
                            if (lineNb == 0) {
                                // Log.v("eXport-itStreamClientS","null line found, trying to read response!!!"); 
                            }
                            break; 
                        }
                        try {
                            line = br.readLine();
                            if (line != null) {
                               line_length = line.length();
                            }
                        } catch (SocketException sex) {
                           break;
                        } catch (IllegalStateException isex) { 
                           // Log.v("eXport-itStreamClientS","IllegalStateException reading following Header Lines of response: "+isex);
                           if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                           }
                           return null;   
                        } catch (IOException ex) {
                           // Log.v("eXport-itStreamClientS","IOexception reading following Header Lines of response: "+ex);
                        } catch (Exception ex) {
                           // Log.v("eXport-itStreamClientS","Exception reading following Header Lines of response: "+ex); 
                           if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                           }
                           return null;
                        }    
                    }
 
                    if ((headers != null) && (!connectionFound)) {
                         headers.add("Connection", "close");
                    }
                    line = null;
                    try {
                        line = br.readLine();
                        if (line != null) {
                            line_length = line.length();
                            rlen = rlen + line_length;
                        }
                    } catch (SocketException sex) {
                       // Log.v("eXport-itStreamClientS","SocketException reading initial Body line of response: "+sex);
                    } catch (IllegalStateException isex) { 
                      // Log.v("eXport-itStreamClientS","IllegalStateException reading initial Body line of response: "+isex);
                      if (in != null) {
                          try { in.close(); } catch (IOException ioe) {  }
                      }
                      return null;   
                    } catch (IOException ex) {
                      // Log.v("eXport-itStreamClientS","IOexception reading initial Body line of response: "+ex);
                    } catch (Exception ex) {
                      // Log.v("eXport-itStreamClientS","exception reading initial Body line of response: "+ex);
                      if (in != null) {
                          try { in.close(); } catch (IOException ioe) {  }
                      }
                      return null;
                    }
                    if (responseContentLength <= 0) {
                        responseContentLength = 65000000;
                    }
                    // Log.v("eXport-itStreamClientS","receiving content length: "+responseContentLength);
                    body_length = 0;
                    bodyresult = "";
                    r = 0;
                    while (line != null) {  // ((responseContentLength > body_length) && 
                        bodyresult = bodyresult + line + "\r\n";
                        body_length = body_length + line.length() + 2;
                        try {
                            line = br.readLine();
                        } catch (SocketException sex) {
                           // Log.v("eXport-itStreamClientS","SocketException reading next Body line of response: "+sex);
                        } catch (IllegalStateException isex) { 
                           // Log.v("eXport-itStreamClientS","IllegalStateException reading next Body line of response: "+isex);
                           if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                           }
                           return null;   
                        } catch (IOException ex) {
                           // Log.v("eXport-itStreamClientS","IOexception reading next Body line of response: "+ex);
                        } catch (Exception ex) {
                           // Log.v("eXport-itStreamClientS","Exception reading next Body line of response: "+ex);
                           if (in != null) {
                               try { in.close(); } catch (IOException ioe) {  }
                           }
                           return null;
                        }
                        // if (bodyresult != null) { body_length = bodyresult.length(); }
                    }
                    if (bodyresult != null) { body_length = bodyresult.length();
                    } else { body_length = 0; }
                    // Log.v("eXport-itStreamClientS","received body length: "+body_length+"  contentlength= "+responseContentLength);
                } else {
                    if (in != null) {
                        try { in.close(); } catch (IOException ioe) {
                            // Log.v("eXport-itStreamClientS","closing input stream exception: "+ioe);
                        }
                    }
                    // Log.v("eXport-itStreamClientS","receiving response ==> br is NULL ");
                    return null;
                }
            
            if (bodyresult.length() > 2) {
                if (httpName.toUpperCase().equals("POST")) {
                    int lng = 0;
                    try {
                        byte[] BodyByte = bodyresult.getBytes("UTF-8");
                        lng = BodyByte.length;
                    } catch (UnsupportedEncodingException e) {
                    }
                    String result = bodyresult;
                    boolean modified = false;
                    if (result != null) {
                       /* if (result.indexOf("&lt;") > 0) {
                          result = result.replace("&lt;","<");
                          modified = true;
                        }    
                        if (result.indexOf("&gt;") > 0) {
                           // Log.v("eXport-itStreamClientS"," result contains &gt; "+result);
                           result = result.replace("&gt;",">");
                           modified = true;
                        } */
                        if (result.indexOf("&quot;") > 0) {
                            result = result.replace("&quot;","\"");
                            modified = true;
                        }
                        if ((result != null) && (modified)) {
                             bodyresult = result;
                        }
                    }
                    // Log.v("eXport-itStreamClientS","POST response Body length="+bodyresult.length()+"/"+lng+" "+bodyresult);
                }

                responseMessage = new StreamResponseMessage(bodyresult, new ContentTypeHeader(contentType));
                if (responseMessage != null) {
                    UpnpHeaders upnpHeaders2 = new UpnpHeaders(); 
                    if ((headers != null) && (upnpHeaders2 != null)) {
                        for (Map.Entry<String, List<String>> entry : headers.entrySet()) {
                            for (String value : entry.getValue())  {
                                upnpHeaders2.add(entry.getKey(), value);
                                /* if (entry.getKey().equals(UpnpHeader.Type.APPLICATIONURL)) {
                                    Log.v("eXport-it StreamClient"," ApplicationURL header added to upnpHeader2!!! "+applicationUrl);
                                } */
                            }
                        }
                    }
                    if (!upnpHeaders2.containsKey(UpnpHeader.Type.HOST)) {
                        upnpHeaders2.add(UpnpHeader.Type.HOST, new HostHeader(host, port));
                    }
                    if (!upnpHeaders2.containsKey(UpnpHeader.Type.EXT)) {
                        upnpHeaders2.add(UpnpHeader.Type.EXT, new EXTHeader());
                    }
                    if (!upnpHeaders2.containsKey(UpnpHeader.Type.USER_AGENT)) {
                        UserAgentHeader userAgentHdr = new UserAgentHeader();
                        userAgentHdr.setString("user-agent: Linux/3.18.91-14843133-QB28034466 UPnP/1.0 DLNADOC/1.50 eXport-it/1.0");
                        upnpHeaders2.add(UpnpHeader.Type.USER_AGENT,userAgentHdr);
                    }
                    if (applicationUrl.length() > 1) {
                        // Log.v("eXport-it StreamClient","response message setting ApplicationURL: "+applicationUrl);
                        ApplicationURLHeader applUrlHdr = new ApplicationURLHeader();
                        applUrlHdr.setString(applicationUrl);
                        upnpHeaders2.add(UpnpHeader.Type.APPLICATIONURL, applUrlHdr); 
                    }
                    responseMessage.setHeaders(upnpHeaders2);
                } else {
                    // Log.v("eXport-it StreamClientS","Stream response message is NULL!!!!");
                    return null;
                }
            } else {
                // Log.v("eXport-it StreamClientS","getting Response message with empty body!!! "+bodyresult2);
                if (bodyresult2.length() < 2) {
                   return null;
                }
            }

        try {    
            br.close();
            in.close();
            // theOutput.close();
        } catch (SocketException sex) {
            // Log.v("eXport-itStreamClientS","SocketException reading response: "+sex);
            return null;  
        } catch (IllegalStateException isex) { 
            // Log.v("eXport-itStreamClientS","IllegalStateException reading response: "+isex);
            return null;   
        } catch (IOException ex) {
            // Log.v("eXport-itStreamClientS","IOexception reading response: "+ex);
            return null;    
        } catch (Exception ex) {
            // Log.v("eXport-itStreamClientS","exception reading response: "+ex);
            return null;
        }    

        if (clientSocket != null) {
            if (!clientSocket.isClosed()) {
                try {
                     clientSocket.close();
                 } catch (Exception ex) { }
            }
        }
        // Date end = new Date();
        // strTime = sdfTime.format(end);
        // Log.v("eXport-it StreamClient",strTime+" SendRequest end message " + requestMessage.toString()+" response body length: "+body_length);
        // Log.v("eXport-it StreamClient","returning Response message body length: "+bodyresult.length());
        return responseMessage;
        
    }
    
    @Override
    public void stop() {
        // Log.v("eXport-it StreamClient","Shutting down HTTP client connection manager/pool");
        stopped = true;

    }
    
    private Socket openSocket(String server, int port) throws Exception {
       Socket socket = null;
    
       // create a socket with a timeout
       try {
           InetAddress inteAddress = InetAddress.getByName(server);
           SocketAddress socketAddress = new InetSocketAddress(inteAddress, port);
  
           // create a socket
           socket = new Socket();
  
           // this method will block no more than timeout ms.
           int timeoutInMs = connectionTimeoutSeconds*1000;   // 10 seconds
           int readTimeOut = dataReadTimeoutSeconds*1000; 
           int size =  socketbuffersize;
           socket.setSoTimeout(readTimeOut);
           socket.setReceiveBufferSize(size);
           socket.setSendBufferSize(size);
           socket.connect(socketAddress, timeoutInMs);
      
           
       } catch (SocketTimeoutException ste) {
      
       }
       return socket;
    }
    
    protected void printHeader(PrintWriter pw, String key, String value) {
        pw.append(key).append(": ").append(value).append("\r\n");
    }

    /**
     * Find byte index separating header from body.
     * It must be the last byte of the first two sequential new lines.
     **/
    private int findHeaderEnd(final byte[] buf, int rlen) {
	int splitbyte = 0;
	while (splitbyte + 3 < rlen) 	{
		if (buf[splitbyte] == '\r' && buf[splitbyte + 1] == '\n' && buf[splitbyte + 2] == '\r' && buf[splitbyte + 3] == '\n') {
			return splitbyte + 4;
		}	
		splitbyte++;
	}
	return 0;
    }
    
}
