/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Copyright (C) March 2020 Design & Development of Computerized Solutions, Belgium
 * program modified by Philippe Gillain for eXport-it application, removing Apache HTTP Client library requirement.
 *
 */

package org.teleal.cling.transport.impl;

// import android.util.Log;

import org.teleal.cling.transport.Router;
import org.teleal.cling.transport.spi.InitializationException;
import org.teleal.cling.transport.spi.StreamServer;
import org.teleal.cling.model.message.StreamRequestMessage;
import org.teleal.cling.model.message.StreamResponseMessage;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.concurrent.Executor;

/**
 *
 * @author Christian Bauer
 */
public class StreamServerImpl implements StreamServer {

    final protected StreamServerConfigurationImpl configuration;

    private static Router router;
    private ServerSocket serverSocket;
    private volatile boolean stopped = false;
    private static int StreamServerPort=0;
    private static int SendBufferSize = 32768;
    private Socket mySocket;
    private int SO_TIMEOUT = 30000;
    private Executor executor = null;
    
    public StreamServerImpl(StreamServerConfigurationImpl configuration) {
        this.configuration = configuration;
        router = this.configuration.getRouter();
        try {
            init(router);
        } catch (InitializationException iex) {
            // Log.v("eXport-it StreamServer","Initialization exception "+iex.toString());
        }
    }

    public StreamServerConfigurationImpl getConfiguration() {
        return configuration;
    }

    // synchronized public void init(InetAddress bindAddress, Router router) throws InitializationException {
    @Override
    public void init(final Router router) throws InitializationException {     // synchronized
            this.router = router;

            // SocketAddress addr = new InetSocketAddress(bindAddress,configuration.getListenPort());
            StreamServerPort = configuration.getListenPort();
            // this.serverSocket = new ServerSocket(
            //                 configuration.getListenPort(),
            //                 configuration.getTcpConnectionBacklog(),
            //                 bindAddress);
            
            // StreamServerPort = serverSocket.getLocalPort();

            SO_TIMEOUT = configuration.getDataWaitTimeoutSeconds() * 1000;
            int SOCKET_BUFFER_SIZE = configuration.getBufferSizeKilobytes() * 1024;
            boolean STALE_CONNECTION_CHECK = configuration.isStaleConnectionCheck();
            boolean TCP_NODELAY = configuration.isTcpNoDelay();
             
             SendBufferSize = configuration.getBufferSizeKilobytes() * 1024;
             // Log.v("eXport-it StreamServer","client buffersize="+configuration.getBufferSizeKilobytes()+" KB  NoDelay:"+configuration.isTcpNoDelay());
            executor = this.router.getSyncProtocolExecutor();
    }
    
    @Override
    public void run () {
        // Log.v("eXport-it StreamServer","starting run process");
        try {
            serverSocket = new ServerSocket(StreamServerPort);  // StreamServerPort,configuration.getTcpConnectionBacklog()
        } catch (IOException ex) {
            // throw new InitializationException("Could not initialize "+getClass().getSimpleName()+": " + ex.toString(), ex);
            // Log.v("eXport-it StreamServer","StreamServerSocket exception: "+ ex.toString());
        }
        if (serverSocket != null) {
          try {
            serverSocket.setReuseAddress(true);
          } catch (SocketException ex) {
            // Log.v("eXport-it StreamServer", "Server Socket setReuseAddress SocketException " + ex);
          }            
          InetAddress address = null;
          SocketAddress addr = null;
          try {
            address = InetAddress.getByName("::");
          } catch (UnknownHostException ex) {
            
          } 
          /* if (address != null) {
            addr = new InetSocketAddress(address, StreamServerPort);
          } else { 
            addr = new InetSocketAddress("0.0.0.0", StreamServerPort); 
          } 
          try {
            serverSocket.bind(addr);
          } catch (IOException ex) {
             Log.v("eXport-it StreamServer", "Server Socket bind IOException " + ex);
          } catch (Exception ex) {
             Log.v("eXport-it StreamServer", "Server Socket bind Exception " + ex); 
          } */ 
          try {
             serverSocket.setReceiveBufferSize(SendBufferSize);
          } catch (SocketException ex) {
             // Log.v("eXport-it StreamServer", "Server Socket setReceiveBufferSize SocketException " + ex);
          }   
          // Log.v("eXport-it StreamServer","Entering blocking receiving loop, listening for HTTP stream requests on: " + serverSocket.getLocalSocketAddress());        
          while (true) {
                Socket s = null;  
                try {  
                   s = serverSocket.accept(); 
                } catch (IOException ex) {
                   // Log.v("eXport-it StreamServer","IO error on server socket, stopping service: " + ex.getMessage()); 
                   s = null;
                   break;
                }
                if (s != null) {
                    try {
                        s.setSoTimeout(SO_TIMEOUT);
                        s.setSoLinger(false,0);
                        s.setTcpNoDelay(false);
                        s.setReceiveBufferSize(SendBufferSize);
                        s.setSendBufferSize(SendBufferSize);
                    } catch (SocketException sox) {
                    } catch (IllegalStateException ex) { 
                    } catch (Exception ex) {    
                    }
                   // Log.v("eXport-it StreamServer"," starting ASync Execution of HttpSession");  
                   handleAsync(s);
                }
          }
          try {
              //log.fine("Receiving loop stopped");
            if (!serverSocket.isClosed()) {
              // log.fine("Closing streaming server socket");
              serverSocket.close();
            }
          } catch (Exception ex) {
            // Log.v("eXport-it StreamServer","Exception closing streaming server socket: " + ex.getMessage());
          }       
        }
    }
    
    @Override
    public int getPort() {
        return StreamServerPort;
    }

    @Override
    public void stop() {
        // Log.v("eXport-it StreamServer","receiving stop command");
        stopped = true;
        if (serverSocket != null) {
          try {
            serverSocket.close();
          } catch (IOException ex) {
            // Log.v("eXport-it StreamServer","Exception closing streaming server socket: " + ex);
          }
        }
    }

    private void handleAsync(final Socket socket) {
        if (executor != null) {
            executor.execute(new HttpSession(router, socket));
        } else {
            HttpSession hs = new HttpSession(router, socket);
            Thread t = new Thread(hs);
            t.setDaemon(false);
            t.start();
        }
    }

}